reset()

from sage.misc.sage_timeit import sage_timeit

import numpy
from numpy import linalg
numpy.random.seed(60856)

import scipy
from scipy.stats.mstats import mquantiles

import matplotlib
matplotlib.use('SVG')
matplotlib.rc('font',family='M+ 2p',weight='light')
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse

from sage.misc.decorators import infix_operator
@infix_operator('multiply')
def dot(a,b) : return vector([a[i]*b[i] for i in range(0,min(len(a),len(b)))])

print(numpy.random.randint(10000))
if (numpy.random.randint(10000) != 7990) : print("Warning: Your random number generator did not generate the same number as mine, results may vary")

print("A Fading Example, track the amplitude of a Rayleigh fading")

factors = [0.5 for _ in range(0,3)]
def f(x,w) : return vector( [1.0*x[0]] + [ factors[k]*x[k+1] + w[k] for k in range(0,len(factors))])

var('theta')
xvars = ['x'+str(k) for k in range(0,len(factors))]
var(reduce(lambda u,v : u+" "+v,xvars))
xvars = [theta]+map(eval,xvars)

wvars = ['w'+str(k) for k in range(0,len(factors))]
var(reduce(lambda u,v : u+" "+v,wvars))
wvars = map(eval,wvars)

fe = f(xvars,wvars)
F = matrix([ [diff(fe[j],xvars[k]) for k in range(0,len(xvars))] for j in range(0,len(xvars))]).subs({wvars[k]:0 for k in range(0,len(wvars))})
L = matrix([ [diff(fe[j],wvars[k]) for k in range(0,len(wvars))] for j in range(0,len(xvars))])

#print(sage_timeit("H.subs({xvars[k]:lp[k] for k in range(0,len(xvars))})", globals(),preparse=False,number=50))
#print(sage_timeit("He(lp)", globals(),preparse=False,number=50))


Q_root = diagonal_matrix([0.2 for _ in range(0,len(wvars))])
Q = Q_root*Q_root.transpose()

def h(x,v) : return vector( [ (x[k+1])*(x[0]) + v[k] for k in range(0,len(xvars)-1)])

var('t')
vvars = ['v' + str(k) for k in range(0,len(xvars)-1)]
var(reduce(lambda u,v : u+" "+v,vvars))
vvars = map(eval,vvars)

dt = 0.2
he = h(xvars,vvars)
H = matrix([ [diff(he[j],xvars[k]) for k in range(0,len(xvars))] for j in range(0,len(vvars))]).subs({vvars[k]:0 for k in range(0,len(vvars))})
J = matrix([ [diff(he[j],vvars[k]) for k in range(0,len(vvars))] for j in range(0,len(vvars))])

Hee_s = str(H.list())
Hee_s = reduce(lambda u,v : u.replace(v[0],v[1]),[(str(xvars[k]),'x['+str(k)+']') for k in range(0,len(xvars))],Hee_s)
Hee_s = 'matrix('+str(len(vvars))+','+str(len(xvars))+','+Hee_s+')'

Hee_e = 'def Hee(x) : return '+Hee_s
exec Hee_e

Jee_s = str(J.list())
Jee_s = reduce(lambda u,v : u.replace(v[0],v[1]),[(str(xvars[k]),'x['+str(k)+']') for k in range(0,len(xvars))],Jee_s)
Jee_s = 'matrix('+str(len(vvars))+','+str(len(vvars))+','+Jee_s+')'

Jee_e = 'def Jee(x) : return '+Jee_s
exec Jee_e

R_root = diagonal_matrix([0.05 for _ in range(0,len(vvars))])
R = R_root*R_root.transpose()
def obs(x,v) : return H*x+J*v

npoints = 200
num_steps = 20

s0h = vector([0.0]+(Q_root*vector(numpy.random.normal(0.3,1,len(factors)))).list())
#s0h = vector([0.0 for _ in range(0,len(xvars))])

P0_root = diagonal_matrix( [2]+Q_root.diagonal())
P0 = P0_root*P0_root.transpose()

#Define the initial state mean
print("The initial state")
print("    Mean")
print(s0h.n(digits=3))
print("    Cov")
print(P0.n(digits=3))
print("F")
print(F)

def create_realization() : 
    s0 = s0h + P0_root*vector(numpy.random.normal(0,1,len(xvars)))
    w_noise = [Q_root*vector(numpy.random.normal(0,1,len(wvars))) for k in range(0,num_steps)]
    v_noise = [R_root*vector(numpy.random.normal(0,1,len(vvars))) for k in range(0,num_steps)]
    truth_state = [s0 for k in range(0,num_steps+1)]
    time = [0 for k in range(0,num_steps+1)]
    for k in range(1,num_steps+1) :
        truth_state[k] = f(truth_state[k-1],w_noise[k-1])
        time[k] = time[k-1]+dt
    z = [h(truth_state[k],v_noise[k-1]) for k in range(1,num_steps+1)]
    return (truth_state,w_noise,v_noise,time,z)

realizations = [create_realization() for _ in range(0,npoints)]

load("Generate_Sigma_Points.sage")
#import Generate_Sigma_points

def gen_set(gen) : return ( gen(len(xvars)).n(), gen(len(wvars)).n(), gen(len(vvars)).n() )

Simplex_set = gen_set(Generate_Simplex)
Order3_nopair = gen_set(Generate_Order3_nopair)
Order5_nopair = gen_set(Generate_Order5_nopair)
Order7_nopair = gen_set(Generate_Order7_nopair)
Order9_nopair = gen_set(Generate_Order9_nopair)
Order4_pair = gen_set(Generate_Order4_pair)

spoint_sets = [
#    (Simplex_set,"UKF_simp","UKF (Simp)"),
#    (Order3_nopair,"UKF_O3","UKF (O3)","#00a250"),
    (Order5_nopair,"UKF_O5","UKF (O5)","#336633"),
    (Order7_nopair,"UKF_O7","UKF (O7)","#00a9e4"),
    (Order9_nopair,"UKF_O9","UKF (O9)","#336699"),
    (Order4_pair,"UKF_O4p","UKF (O5f)","#ef5f21")
]

def run_linearized(ss0h,zz,lin_points) :
    ss = [ss0h for k in range(0,num_steps+1)]
    pp = [P0 for k in range(0,num_steps+1)]
    for k in range(1,num_steps+1) :
        s = ss[k-1]
        p = pp[k-1]
        lp = lin_points[k-1]
 #       sub_point = {xvars[k]:lp[k] for k in range(0,len(xvars))}
        Fe = F
        Le = L
        sn = f(lp,[0 for _ in range(0,len(wvars))])+Fe*(s-lp)
        pn = Fe*p*Fe.transpose()+Le*Q*Le.transpose()
        He = Hee(lp)
        Je = Jee(lp)
        K = pn*He.transpose()*(He*pn*He.transpose()+Je*R*Je.transpose()).inverse()
        tr = zz[k-1]-(h(lp,vector([0 for _ in range(0,len(vvars))]))+He*(sn-lp))#We can do this because obs is linear
        su = sn + K*tr
        tt = identity_matrix(len(xvars))-K*He
        pu = tt*pn*tt.transpose()+K*Je*R*Je.transpose()*K.transpose()
        ss[k] = su
        pp[k] = pu
    return(ss,pp)

def run_ekf(ss0h,zz) :
    ss = [ss0h for k in range(0,num_steps+1)]
    pp = [P0 for k in range(0,num_steps+1)]
    for k in range(1,num_steps+1) :
        s = ss[k-1]
        p = pp[k-1]
        lp = s
#        sub_point = {xvars[k]:lp[k] for k in range(0,len(xvars))}
        Fe = F
        Le = L
        sn = f(lp,[0 for _ in range(0,len(wvars))])+Fe*(s-lp)
        pn = Fe*p*Fe.transpose()+Le*Q*Le.transpose()
        lp = sn
#        sub_point = {xvars[k]:lp[k] for k in range(0,len(xvars))}
        He = Hee(lp)
        Je = Jee(lp)
        K = pn*He.transpose()*(He*pn*He.transpose()+Je*R*Je.transpose()).inverse()
        tr = zz[k-1]-h(sn,vector([0 for _ in range(0,len(vvars))])) #We can do this because obs is linear
        su = sn + K*tr
        tt = identity_matrix(len(xvars))-K*He
        pu = tt*pn*tt.transpose()+K*Je*R*Je.transpose()*K.transpose()
        ss[k] = su
        pp[k] = pu
    return(ss,pp)

def eval_spoint(ef,s,p_r,v,r_r,spoints1,spoints2) :
    weights1 = spoints1[0]
    spoints_t1 = spoints1.submatrix(1,0).transpose()
    y0 = ef(s,v)
    tss = [s+p_r*spoints_t1[j] for j in range(0,len(weights1))]
    tys = [ef(tss[j],v) for j in range(0,len(weights1))]
    weights2 = spoints2[0]
    spoints_t2 = spoints2.submatrix(1,0).transpose()
    tvv = [v+r_r*spoints_t2[j] for j in range(0,len(weights2))]
    tyv = [ef(s,tvv[j]) for j in range(0,len(weights2))]
    tyv = [y0 for _ in range(0,len(weights2))]
    ym = -y0+sum([weights1[j]*tys[j] for j in range(0,len(weights1))]) + sum([weights2[j]*tyv[j] for j in range(0,len(weights2))])
    pzz = -numpy.outer((y0-ym),(y0-ym)) # this is so that we can merge the two sets
    pzz = sum([weights1[j]*numpy.outer(tys[j]-ym,tys[j]-ym) for j in range(0,len(weights1))])
    pzz = pzz + sum([weights2[j]*numpy.outer(tyv[j]-ym,tyv[j]-ym) for j in range(0,len(weights2))])
    pxz = sum([weights1[j]*numpy.outer(tss[j]-s,tys[j]-ym) for j in range(0,len(weights1))])
    return (ym,matrix(pzz),matrix(pxz))

def run_spoint(ss0h,zz,spoints) :
    ss = [ss0h for k in range(0,num_steps+1)]
    pp = [P0 for k in range(0,num_steps+1)]
    for k in range(1,num_steps+1) :
        s = ss[k-1]
        p = pp[k-1]
        sn = F*s
        pn = F*p*F.transpose() + L*Q*L.transpose()
#        tmp = linalg.svd(p)
#        p_root = matrix(tmp[0])*diagonal_matrix(vector(tmp[1]).apply_map(sqrt))*matrix(tmp[2])
#        (sn,pn,pnp) = eval_spoint(f,s,p_root,vector([0 for _ in range(0,len(wvars))]),Q_root,spoints[0],spoints[1])
        tmp = linalg.svd(pn)
        pn_root = matrix(tmp[0])*diagonal_matrix(vector(tmp[1]).apply_map(sqrt))*matrix(tmp[2])
        (y,pyy,pxy) = eval_spoint(h,sn,pn_root,vector([0 for _ in range(0,len(vvars))]),R_root,spoints[0],spoints[2])
        K = pxy*(pyy).inverse()
        tr = zz[k-1]-y #We can do this because obs is linear
        su = sn + K*tr
        pu = pn - K*pyy*K.transpose()
        ss[k] = su
        pp[k] = pu
    return(ss,pp)

def generate_xt(time,truth,s_fil_p,s_fil,name,col,filename,full_label) :
    fig = plt.figure()
    fig.set_figwidth(3)
    fig.set_figheight(2.5)
    axs = fig.add_axes( (0.6/7,0.4/4,1-.8/7,1-.6/4))

    axs.set_xlabel("Time")
    axs.set_ylabel("Theta")

    tru, = axs.plot(time,map(lambda u : u[1],truth),'k-',linewidth=3.5)
    per, = axs.plot(time,map(lambda u : u[0],s_fil_p),'^-',color='#8F00FF',linewidth=.7,markeredgecolor='k',markeredgewidth=0.7,markersize=6)
    fil, = axs.plot(time,map(lambda u : u[0],s_fil),'v--',color=col,linewidth=1.5,markeredgecolor='k',markersize=5,markeredgewidth=0.5)

    if full_label : 
        axs.legend( (tru,per,fil), ("Truth","Perfect (KF)",name), framealpha=0.7,numpoints=1,prop={'family':'M+ 2p','weight':'ultralight','size':8})
    else : 
        axs.legend( [fil], [name], framealpha=0.7,numpoints=1,prop={'family':'M+ 2p','weight':'ultralight','size':8})

    yticks = [0,1,-1]
    yticks.sort()
    axs.set_yticks(yticks)
    axs.set_yticklabels(map(lambda u : u.n(digits=2),yticks), fontsize=8, family='sans-serif', weight='light')
    axs.set_xticks([])
    #axs.set_xticks(time,minor=True)
    axs.grid(True)

    fig.savefig('Plots/Fadeing_Time_'+filename+'.svg')
    plt.close(fig)

#Run the first set
this_run = realizations[0]
time = this_run[3]
(s_per,p_per) = run_linearized(s0h,this_run[4],this_run[0])
(s_ekf,p_ekf) = run_ekf(s0h,this_run[4])

#generate_xt(this_run[3],this_run[0],s_per,s_ekf,"EKF",clist[0],"EKF",True)
for k in range(0,len(spoint_sets)) :
    (s_sp,p_sp) = run_spoint(s0h,this_run[4],spoint_sets[k][0])
#    generate_xt(this_run[3],this_run[0],s_per,s_sp,spoint_sets[k][2],clist[k+1],spoint_sets[k][1],False)

err_per = [ [] for k in range(0,npoints)]
err_ekf = [ [] for k in range(0,npoints)]
err_sp = [ [ [] for k in range(0,npoints)] for j in range(0,len(spoint_sets))]

def find_err(truth,fil) :
    d_err = [truth[k][0] - fil[k][0] for k in range(0,len(fil))]
    return map(lambda u : sqrt(u*u), d_err)


@parallel
def perform_run(this_run) :
    (s_per,p_per) = run_linearized(s0h,this_run[4],this_run[0])
    terr_per = find_err(this_run[0],s_per)
    (s_ekf,p_ekf) = run_ekf(s0h,this_run[4])
    terr_ekf = find_err(this_run[0],s_ekf)
    terr_sp = [ [] for k in range(0,len(spoint_sets)) ]
    for k in range(0,len(spoint_sets)) :
        (s_sp,p_sp) = run_spoint(s0h,this_run[4],spoint_sets[k][0])
        terr_sp[k] = find_err(this_run[0],s_sp)
    return (terr_per,terr_ekf,terr_sp)

runs = list(perform_run([(realizations[j],) for j in range(0,npoints)]))

err_per = [ runs[k][1][0] for k in range(0,npoints) ]
err_ekf = [ runs[k][1][1] for k in range(0,npoints) ]
err_sp = [ [runs[k][1][2][j] for k in range(0,npoints)] for j in range(0,len(spoint_sets)) ]
#err_sp = [ runs[k][1][2][0] for k in range(0,npoints)]


def add_plot(q,td,col,axs,sym) :
    tt = map(lambda u : u+td,time)
    qp = axs.plot(tt,list(q),sym,color=col,linewidth=.7,markeredgecolor='k',markeredgewidth=0.7,markersize=6)
    return(qp)

def draw_quartile(qt,name,ybnd,ytk) :
    quant_per = mquantiles(err_per,prob=[qt],axis=0)[0]
    quant_ekf = mquantiles(err_ekf,prob=[qt],axis=0)[0]
#    quant_sp = mquantiles(err_sp,prob=[qt],axis=0)[0]
    quant_sp = [ mquantiles(err_sp[k],prob=[qt],axis=0)[0] for k in range(0,len(spoint_sets))]

    fig = plt.figure()
    fig.set_figwidth(7)
    fig.set_figheight(5)
    axs = fig.add_axes( (0.6/7,0.4/4,1-.8/7,1-.6/4) )

    axs.set_xlabel("Time")
    axs.set_ylabel("Error")

    per, = add_plot(quant_per,0,'#8F00FF',axs,'v-')
    ekf, = add_plot(quant_ekf,0.0,'#cc0000',axs,'v-')
    
    leg_entries = [ per for k in range(0,len(spoint_sets)+2)]
    leg_names = ["" for k in range(0,len(spoint_sets)+2)]
    leg_entries[0] = per
    leg_names[0] = "Perfect (KF)"
    leg_entries[1] = ekf
    leg_names[1] = "EKF"

    for k in range(0,len(spoint_sets)) :
        sp, = add_plot(quant_sp[k],0.0,spoint_sets[k][3],axs,'v-')
        leg_entries[k+2] = sp
        leg_names[k+2] = spoint_sets[k][2]
    axs.legend( tuple(leg_entries) , tuple(leg_names),
                framealpha=0.7,numpoints=1,prop={'family':'M+ 2p','weight':'ultralight','size':10})
#    add_plot(quant_sp,0.0,clist[1],axs,'v-')
    axs.set_xticks( map(lambda u : u*dt,[0,5,10,15,20]) )
    yticks = ytk #+ [quant_ekf[num_steps],quant_per[num_steps],quant_sp[2][num_steps],quant_sp[1][num_steps],quant_sp[0][num_steps]]
    yticks.sort()
    axs.set_yticks( map(lambda u : n(u,digits=2),yticks) )
    axs.set_ylim( ybnd )
    axs.grid(True)
    
    fig.savefig('Plots/Fading_Error_'+name+'.svg')
    plt.close(fig)

draw_quartile(.5,'50th',(0,5),[0,1,2,3,4,5])
draw_quartile(.75,'75th',(0,10),[0,2.5,5,7.5,10])
#draw_quartile(1,'max')
