import numpy
from numpy import linalg

def performAdvancement( Xr, Wr, adv ) :
    xn = adv[0]+adv[1]*Xr[0]+adv[2]*Wr[0]
    pn = adv[1]*Xr[1]*adv[1].transpose()+adv[2]*Wr[1]*adv[2].transpose()+adv[3]
    return (xn,pn)

def performMeasurement(Xr, Vr, meas ) :
    K = Xr[1]*meas[1].transpose()*( meas[1]*Xr[1]*meas[1].transpose()+meas[2]*Vr[1]*meas[2].transpose()+meas[3]).inverse()
    xn = Xr[0] - K * (meas[0] + meas[1]*Xr[0]+meas[2]*Vr[0])
    tf = (identity_matrix(len(Xr[0]))-K*meas[1])
    pn = tf*Xr[1]*tf.transpose()+K*(meas[2]*Vr[1]*meas[2].transpose()+meas[3])*K.transpose()
    return(xn,pn)

def ekfAdvance(f,F,L,linp) :
    xnl = f(linp[0][0],linp[1][0],linp[2]).n()
    tF = F(linp[0][0],linp[1][0],linp[2]).n()
    tL = L(linp[0][0],linp[1][0],linp[2]).n()
    
    xn = xnl - tF*(linp[0][0]) - tL*(linp[1][0])
    
    return (xn,tF,tL,matrix(len(xn),len(xn),0))

def ekfMeasure(h,H,V,linp) :
    ynl = h(linp[0][0],linp[1][0],linp[2]).n()
    tH = H(linp[0][0],linp[1][0],linp[2]).n()
    tV = V(linp[0][0],linp[1][0],linp[2]).n()

    yn = ynl - tH*(linp[0][0]) -tV*(linp[1][0])

    return (yn,tH,tV,matrix(len(yn),len(yn),0))


def sigmaAdvance(f,linp,spoints,alpha) :
    weights = spoints[0]
    spoints_t = spoints.submatrix(1,0).transpose()
    gstate = vector( linp[0][0].list() + linp[1][0].list() )
    gcov = linp[0][1].block_sum(linp[1][1])
    tmp = linalg.svd(gcov)
    gcov_root = matrix(tmp[0])*diagonal_matrix(vector(tmp[1]).apply_map(sqrt))*matrix(tmp[2])
    npoints = len(weights)
    tx = [gstate + alpha*gcov_root*spoints_t[j] for j in range(0,npoints)]
    ts = [tx[j][0:len(linp[0][0])] for j in range(0,npoints)]
    tn = [tx[j][len(linp[0][0]):len(gstate)] for j in range(0,npoints)]
    ty = [f(ts[k],tn[k],linp[2]).n() for k in range(0,npoints)]

    ym = sum([weights[k]*ty[k] for k in range(0,npoints)])

    pyy = matrix(sum([ weights[k]*numpy.outer(ty[k]-ym,ty[k]-ym) for k in range(0,npoints) ])) / (alpha*alpha)
    psy = matrix(sum([ weights[k]*numpy.outer(ts[k]-linp[0][0],ty[k]-ym) for k in range(0,npoints)])) / (alpha*alpha)
    pny = matrix(sum([ weights[k]*numpy.outer(tn[k]-linp[1][0],ty[k]-ym) for k in range(0,npoints)])) / (alpha*alpha)

    tF = (matrix(linalg.pinv(linp[0][1]))*psy).transpose()
    tL = (matrix(linalg.pinv(linp[1][1]))*pny).transpose()
    nonlin = pyy - tF*linp[0][1]*tF.transpose() - tL*linp[1][1]*tL.transpose()

    return (ym-tF*linp[0][0]-tL*linp[1][0],tF,tL,nonlin)
    
    
def sigmaMeasure(f,linp,spoints,alpha) :
    weights = spoints[0]
    spoints_t = spoints.submatrix(1,0).transpose()
    gstate = vector( linp[0][0].list() + linp[1][0].list() )
    gcov = linp[0][1].block_sum(linp[1][1])
    tmp = linalg.svd(gcov)
    gcov_root = matrix(tmp[0])*diagonal_matrix(vector(tmp[1]).apply_map(sqrt))*matrix(tmp[2])
    npoints = len(weights)
    tx = [gstate + alpha*gcov_root*spoints_t[j] for j in range(0,npoints)]
    ts = [tx[j][0:len(linp[0][0])] for j in range(0,npoints)]
    tn = [tx[j][len(linp[0][0]):len(gstate)] for j in range(0,npoints)]
    ty = [f(ts[k],tn[k],linp[2]).n() for k in range(0,npoints)]

    ym = sum([weights[k]*ty[k] for k in range(0,npoints)])

    pyy = matrix(sum([ weights[k]*numpy.outer(ty[k]-ym,ty[k]-ym) for k in range(0,npoints) ])) / (alpha*alpha)
    psy = matrix(sum([ weights[k]*numpy.outer(ts[k]-linp[0][0],ty[k]-ym) for k in range(0,npoints)])) / (alpha*alpha)
    pny = matrix(sum([ weights[k]*numpy.outer(tn[k]-linp[1][0],ty[k]-ym) for k in range(0,npoints)])) / (alpha*alpha)

    tH = (matrix(linalg.pinv(linp[0][1]))*psy).transpose()
    tV = (matrix(linalg.pinv(linp[1][1]))*pny).transpose()
    nonlin = pyy - tH*linp[0][1]*tH.transpose() - tV*linp[1][1]*tV.transpose()

    return (ym-tH*linp[0][0]-tV*linp[1][0],tH,tV,nonlin)
