import scipy
from scipy import optimize

from sage.misc.decorators import infix_operator
@infix_operator('multiply')
def dot(a,b) : return vector([ a[i]*b[i] for i in range(0,min(len(a),len(b)))])

p = 6

def E(d,k,j) :
    ee = matrix(QQ,d)
    ee[k,j]=1
    return ee

def cosine_matrix(n,j,k,theta) : return (cos(theta)-1)*(E(n,j,j)+E(n,k,k))+sin(theta)*(E(n,j,k)-E(n,k,j)) + matrix.identity(n)

def to_cmatrix(n,pp,th) : return prod([cosine_matrix(n,pp[k][0],pp[k][1],th[k]) for k in range(0,len(pp))])

base_vars = reduce(lambda u,v : u+" "+v,["theta"+str(k) for k in range(0,p-1)])
var(base_vars)
base_vlist = eval(base_vars.replace(' ',','))

temp_order1 = [ [0,k+1] for k in range(0,p-1) ]
temp_order1.reverse()
SP_proto_base = to_cmatrix(p,temp_order1,base_vlist)

print("Prototype Base Form, Dimension : " + str(p-1))
print(SP_proto_base)

print("Base Sigma Points")
w_root = SP_proto_base[0]
w_root_i = w_root.apply_map(lambda u: 1/u)
SP_base = matrix(flatten([[w_root *dot* w_root], [SP_proto_base[k] *dot* w_root_i for k in range(1,p) ]]))

tf_vars = reduce(lambda u,v : u+" "+v,["phi"+str(k) for k in range(0,(p-1)*(p-2)/2)])
var(tf_vars)
tf_vlist = eval(tf_vars.replace(' ',','))

print("Sigma Point Transform Matrix, Dimension : "+str((p-1)*(p-2)/2))
templist2 = [[j,k] for j in range(1,p) for k in range(j+1,p)]
#templist.reverse()
tf = to_cmatrix(p,templist2,tf_vlist)

to_set = [
    [[1,1,1],0],
    [[1,1,2],0],
    [[1,2,2],0],
    [[2,2,2],0],
    [[1,1,1,2],0],
    [[1,1,2,2],1.333],
    [[1,2,2,2],0],
    [[1,1,1,1],2],
    [[2,2,2,2],2]
]

def create_condition(proto,tSP) : return tSP[0]*reduce(lambda u,v : u *dot* v,[tSP[proto[0][k]] for k in range(0,len(proto[0]))]) - proto[1]

all_vars = list(base_vlist+tf_vlist)

check_conds = [create_condition(to_set[k],tf*SP_base) for k in range(0,len(to_set))]

check_conds_p = check_conds + [0 for k in range(0,len(all_vars)-len(to_set))]

def fix_cond(cond) :
    str_v = str(cond)
    for k in range(0,len(all_vars)) :
        str_v = str_v.replace(str(all_vars[k]),"u["+str(k)+"]")
    return str_v.replace('^','**')

error_v = lambda u : map(lambda v : v.subs({all_vars[k]:u[k] for k in range(0,len(all_vars))}), check_conds_p)

exec 'def error_v(u) : return '+fix_cond(check_conds_p)

start_point = [  pi.n()*(2*random()-1) for k in range(0,len(all_vars))]

sol = optimize.leastsq(error_v,start_point)

print("Error in Solution")
print(vector(error_v(sol[0])))

print("Solution")
SP_sol =  (tf*SP_base).subs({all_vars[k]:sol[0][k] for k in range(0,len(all_vars))})
print(SP_sol.n(digits=3))

tSP = SP_sol.__copy__()
w_root = map(sqrt,SP_sol[0])
for k in range(0,p) :
    tSP[k] = tSP[k] *dot* w_root
tSP[0] = w_root

w_root_i = tSP[0].apply_map(lambda u : 1/u)
A1 = tSP*diagonal_matrix(tSP[1] *dot* w_root_i)*tSP.transpose()
A2 = tSP*diagonal_matrix(tSP[2] *dot* w_root_i)*tSP.transpose()

print("A1")
print(A1.n(digits=3).submatrix(1,1,2,2))
print("A2")
print(A2.n(digits=3).submatrix(1,1,2,2))
print("A11")
print( (A1*A1).n(digits=3).submatrix(1,1,2,2) )
print("A22")
print( (A2*A2).n(digits=3).submatrix(1,1,2,2) )

w_root_s = (tSP[0]*(sqrt(1/3).n()*sqrt(2).n())).list()

zlist = [0 for k in range(0,p)]
patchwork = matrix([
    w_root_s      + w_root_s      + w_root_s,
    tSP[1].list() + tSP[1].list() + zlist,
    tSP[2].list() + zlist         + tSP[1].list(),
    zlist         + tSP[2].list() + tSP[2].list()
])/sqrt(2)

w_root_i = patchwork[0].apply_map(lambda u : 1/u)
pA1 = patchwork*diagonal_matrix(patchwork[1] *dot* w_root_i)*patchwork.transpose()
pA2 = patchwork*diagonal_matrix(patchwork[2] *dot* w_root_i)*patchwork.transpose()
pA3 = patchwork*diagonal_matrix(patchwork[3] *dot* w_root_i)*patchwork.transpose()
print("pA1")
print(pA1)
print("pA2")
print(pA2)
print("pA3")
print(pA3)

pA11 = patchwork*diagonal_matrix(patchwork[1] *dot* w_root_i *dot* patchwork[1] *dot* w_root_i)*patchwork.transpose()
pA22 = patchwork*diagonal_matrix(patchwork[2] *dot* w_root_i *dot* patchwork[2] *dot* w_root_i)*patchwork.transpose()
pA33 = patchwork*diagonal_matrix(patchwork[3] *dot* w_root_i *dot* patchwork[3] *dot* w_root_i)*patchwork.transpose()
print("pA11")
print(pA11)
print("pA22")
print(pA22)
print("pA33")
print(pA33)


#Other Set
w_root_s = (tSP[0]*(sqrt(1/6).n()*sqrt(3).n())).list()

zlist = [0 for k in range(0,p)]
patchwork = matrix([
    w_root_s      + w_root_s      + w_root_s      + w_root_s      + w_root_s      + w_root_s,
    tSP[1].list() + tSP[1].list() + tSP[1].list() + zlist         + zlist         + zlist,
    tSP[2].list() + zlist         + zlist         + tSP[1].list() + tSP[1].list() + zlist,
    zlist         + tSP[2].list() + zlist         + tSP[2].list() + zlist         + tSP[1].list(),
    zlist         + zlist         + tSP[2].list() + zlist         + tSP[2].list() + tSP[2].list(),
])/sqrt(3)

w_root_i = patchwork[0].apply_map(lambda u : 1/u)
pA1 = patchwork*diagonal_matrix(patchwork[1] *dot* w_root_i)*patchwork.transpose()
pA2 = patchwork*diagonal_matrix(patchwork[2] *dot* w_root_i)*patchwork.transpose()
pA3 = patchwork*diagonal_matrix(patchwork[3] *dot* w_root_i)*patchwork.transpose()
print("pA1")
print(pA1)
print("pA2")
print(pA2)
print("pA3")
print(pA3)

pA11 = patchwork*diagonal_matrix(patchwork[1] *dot* w_root_i *dot* patchwork[1] *dot* w_root_i)*patchwork.transpose()
pA22 = patchwork*diagonal_matrix(patchwork[2] *dot* w_root_i *dot* patchwork[2] *dot* w_root_i)*patchwork.transpose()
pA33 = patchwork*diagonal_matrix(patchwork[3] *dot* w_root_i *dot* patchwork[3] *dot* w_root_i)*patchwork.transpose()
print("pA11")
print(pA11)
print("pA22")
print(pA22)
print("pA33")
print(pA33)


def ffdiag(Ck_) :
    kl = len(Ck_)
    Ck = Ck_
    for k in range(0,kl) :
        Ck[k] = Ck_[k].__copy__()
    V = identity_matrix(p)
    W = matrix(p,p,0.n())
    for ind in range(0,1) :
        z = matrix(p,p,[sum([Ck[m][j,j]*Ck[m][k,k] for m in range(0,kl)]) for j in range(0,p) for k in range(0,p)])
        print(z.n(digits=3))
        y = matrix(p,p,[sum([(1-kronecker_delta(j,k))*Ck[m][k,k]*Ck[m][j,k] for m in range(0,kl)]) for j in range(0,p) for k in range(0,p) ])
        print(y.n(digits=3))
        for j in range(0,p) :
            for k in range(j,p) :
                W[j,k] = (z[j,k]*y[k,j]-z[j,j]*y[j,k])/(z[j,j]*z[k,k]-z[j,k]*z[j,k])
                W[k,j] = (z[j,k]*y[j,k]-z[k,k]*y[k,j])/(z[k,k]*z[j,j]-z[j,k]*z[j,k])
        for k in range(0,p) : W[k,k] = 0
        print(W.n(digits=3))
        Wfro = sqrt(sum(map(lambda u : u*u,W.list())))
        if Wfro>0.8 :
            W = 0.8/Wfro*W
            tf = identity_matrix(p)+W
            V = tf*V
            for k in range(0,kl) :
                Ck[k] = tf*Ck[k]*tf.transpose()
    return (V,Ck)
            
    
tS = tSP.__copy__()
tS[0] = tSP[0] *dot* tSP[0]
for k in range(1,p) :
    tS[k] = tSP[k] *dot* tSP[0].apply_map(lambda u : 1/u)

tTForm = to_cmatrix(p,templist,[2*pi.n()*random() for _ in range(0,(p-1)*(p-2)/2)])

tS_2 = tTForm*tS

tSP_2 = tS_2.__copy__()
tSP_2[0] = tS_2[0].apply_map(lambda u : sqrt(u))
for k in range(1,p) :
    tSP_2[k] = tS_2[k] *dot* tSP_2[0]

tA1_1 = tSP*diagonal_matrix(tSP[1] *dot* tSP[0].apply_map(lambda u : 1/u))*tSP.transpose()
tA1_2 = tSP_2*diagonal_matrix(tSP_2[1] *dot* tSP_2[0].apply_map(lambda u : 1/u))*tSP_2.transpose()

print("tA1_1")
print(tA1_1.n(digits=3))
print("tA1_2")
print(tA1_2.n(digits=3))
print("TF A1")
tprod_TF = tTForm.tensor_product(tTForm)
hype_tform = sum([tTForm[1,k]*tprod_TF for k in range(1,p)])
print( (hype_tform*vector(tA1_1.list())).n(digits=3) )
