reset()

import numpy
from numpy import linalg
numpy.random.seed(60856)

import scipy
from scipy.stats.mstats import mquantiles

from sage.misc.decorators import infix_operator
@infix_operator('multiply')
def dot(a,b) : return vector([a[i]*b[i] for i in range(0,min(len(a),len(b)))])

if (numpy.random.randint(10000) != 7990) : print("Warning: Your random number generator did not generate the same number as mine, results may vary")

npoints = 10000

def f1(x) : return vector([x[0],x[0]*x[0]])
def f2(x) : return vector([x[0],x[0]*x[0]*x[0]])
def f3(x) : return vector([x[0],x[0]*x[0]*x[0]*x[0]])
def f4(x) : return vector([x[0],x[0]*x[0]*x[0]*x[0]*x[0]])
def f5(x) : return vector([x[0],x[1],x[0]*x[1]])

functions = [f1,f2,f3,f4,f5]

P0_root = identity_matrix(2)

#Define the initial state mean
print("The initial state")
s0h = vector([1,1]).n()
P0 = P0_root*P0_root.transpose()
print("    Mean")
print(s0h.n(digits=3))
print("    Cov")
print(P0.n(digits=3))

def create_realization() : 
    s0 = s0h + P0_root*vector(numpy.random.normal(0,1,2))
    return s0

realizations = [create_realization() for _ in range(0,npoints)]

var('x0,x1')
moment_subs = [x0**10==945,x0**8==105,x0**6==15,x0**4==3,x1**2*x0**2==1,x0**2==1,x1**2==1,x0==0,x1==0]
def apply_moments(x) : return reduce(lambda u,v:u.subs(v),moment_subs,x)

load("Generate_Sigma_Points.sage")

Simplex_set = Generate_Simplex(2).n()
Order5_nopair = Generate_Order5_nopair(2).n()
Order4_pair = Generate_Order4_pair(2).n()
Order9_nopair = Generate_Order9_nopair(2).n()

def run_spoint(s0h,f,spoints) :
    weights = spoints[0]
    spoints_t = spoints.submatrix(1,0).transpose()
    ty = [f(s0h) for _ in range(0,len(weights))]
    for j in range(0,len(weights)) :
        tsp = spoints_t[j]
        ts = s0h + P0_root*vector([tsp[0],tsp[1]])
        ty[j] = f(ts).n()
    ym = sum([weights[j]*ty[j] for j in range(0,len(weights))])
    pzz = sum([ weights[j]*(ty[j]-ym).outer_product(ty[j]-ym) for j in range(0,len(weights))])
    return (ym,pzz)

print("")
print("")

for f_ind in range(0,len(functions)) :
    print("Function: "+str(f_ind+1))
    print(functions[f_ind]([x0,x1]))
    z = functions[f_ind]([x0+s0h[0],x1+s0h[1]])
    print("Actual Expected Value: " + str(apply_moments(z.apply_map(expand)).n(digits=3)))
    c = matrix( [[ expand((z[j]-apply_moments(expand(z[j])))*(z[k]-apply_moments(expand(z[k])))) for j in range(0,len(z))] for k in range(0,len(z))])
    print("Actual Covariance: ")
    print(apply_moments(c).n(digits=3))

    obs = map(functions[f_ind],realizations)
    print("Measured Mean: "+str(vector(numpy.mean(obs,0)).n(digits=3)))
    print("Measured Covariance: ")
    print(matrix(numpy.cov(obs,rowvar=0)).n(digits=3))

    print("Linear Expected Value: "+str(functions[f_ind](s0h).n(digits=3)))
    ye = functions[f_ind](s0h)
    He = matrix( [diff(z,x0),diff(z,x1)]).transpose().subs({x0:s0h[0],x1:s0h[1]})
    print("Linear Covariance: ")
    print( (He*P0*He.transpose()).n(digits=3))

    (ym,pzz) = run_spoint(s0h,functions[f_ind],Simplex_set)
    print("Simplex Mean: "+str(ym.n(digits=3)))
    print("Simplex Covariance: ")
    print(pzz.n(digits=3))

    (ym,pzz) = run_spoint(s0h,functions[f_ind],Order5_nopair)
    print("Order 5 Mean: "+str(ym.n(digits=3)))
    print("Order 5 Covariance: ")
    print(pzz.n(digits=3))
    
    (ym,pzz) = run_spoint(s0h,functions[f_ind],Order4_pair)
    print("Order 4 xross Mean: "+str(ym.n(digits=3)))
    print("Order 4 xross Covariance: ")
    print(pzz.n(digits=3))

    (ym,pzz) = run_spoint(s0h,functions[f_ind],Order9_nopair)
    print("Order 9 Mean: "+str(ym.n(digits=3)))
    print("Order 9 Covariance: ")
    print(pzz.n(digits=3))
    print("")
    print("")
