reset()

import numpy
from numpy import linalg
numpy.random.seed(96097)

import matplotlib
matplotlib.use('SVG')
matplotlib.rc('font',family='M+ 2p',weight='light')
import matplotlib.pyplot as plt

import scipy
from scipy.stats.mstats import mquantiles

from sage.misc.decorators import infix_operator
@infix_operator('multiply')
def dot(a,b) : return vector([a[i]*b[i] for i in range(0,min(len(a),len(b)))])

if (numpy.random.randint(10000) != 3648) : print("Warning: Your random number generator did not generate the same number as mine, results may vary")

npoints = 1000

def h(x,v) : return vector([v[0]])+vector([sum( [(-x[0]/n)^n for n in range(1,5)] )])
def h_1(x,v) : return vector([v[0]])+vector([sum( [(-x[0]/n)^n for n in range(1,2)]) ])
def h_2(x,v) : return vector([v[0]])+vector([sum( [(-x[0]/n)^n for n in range(1,3)]) ])
def h_3(x,v) : return vector([v[0]])+vector([sum( [(-x[0]/n)^n for n in range(1,4)]) ])

def fit_2(r) :
    var('a0 a1 a2')
    var('x')
    hev = h([x],[0])[0]
    tpl = a0+a1*x+a2*x^2
    pd2 = (hev-tpl)^2
    idf = integrate(pd2,(x,-r,r))
    tsl = solve([diff(idf,a0),diff(idf,a1),diff(idf,a2)],[a0,a1,a2],solution_dict=True)
    return lambda u : vector([tpl.subs(tsl[0]).subs({x:u[0]})])

p_res = 20
v1 = [k/p_res for k in range(-8*p_res,8*p_res)]
hv = map(lambda x : h([x],[0])[0],v1)
hv_1 = map(lambda x : h_1([x],[0])[0],v1)
hv_2 = map(lambda x : h_2([x],[0])[0],v1)
hv_3 = map(lambda x : h_3([x],[0])[0],v1)

fig = plt.figure()
fig.set_figwidth(7)
fig.set_figheight(4)
axs = fig.add_axes( ((0.6)/7, (0.4)/4, 1-.8/7, 1-.6/4) )

axs.set_xlabel('x')
axs.set_ylabel('h(x,0)')
axs.plot(v1,hv,'-',color='#cc0000')
axs.plot(v1,hv_1,'-',color='#336633')
axs.plot(v1,hv_2,'-',color='#336699')
axs.plot(v1,hv_3,'-',color='#ff9900')
fig.savefig('Plots/Region_h_x.svg')
plt.close(fig)


var('x0')
moment_subs = [x0**10==945,x0**8==105,x0**6==15,x0**4==3,x0**2==1,x0==0]
def apply_moments(x) : return reduce(lambda u,v:u.subs(v),moment_subs,x)

s0h = vector([0]).n()

R_root = matrix([[1/10]])
R = R_root*R_root.transpose()

def create_realization(P0_root) : 
    s0 = s0h + P0_root*vector([numpy.random.normal(0,1)])
    v = R_root*vector([numpy.random.normal(0,1)])
    return (s0,v)


load("Generate_Sigma_Points.sage")

Simplex_set = Generate_Simplex(1).n()
Order5_nopair = Generate_Order5_nopair(1).n()
Order7_nopair = Generate_Order7_nopair(1).n()
Order9_nopair = Generate_Order9_nopair(1).n()

def max_point(sp_set) :
    spoints_t = sp_set.submatrix(1,0).transpose()
    max = 0
    for k in range(0,len(sp_set[0])) :
        d = spoints_t[k]*spoints_t[k]
        if d>max : max = d
    return d

prtl = (90,60)

#sp_dists = (1.15,0.674)
sp_dists = (n(scipy.special.erfinv(prtl[0]/100)*sqrt(2)),n(scipy.special.erfinv(prtl[1]/100)*sqrt(2)))
#sp_dists = (2,.1)

spoint_sets_unscaled = [
    (Order9_nopair,'O9',1),
    (Order7_nopair,'O7',1),
    (Order5_nopair,'O5',1),
    (Simplex_set,'Simp/O3',1)
]

spoint_sets = [
    (Order7_nopair,'O7 ('+str(prtl[0])+'%)', sp_dists[0]/max(Order7_nopair[1].apply_map(abs))),
    (Order5_nopair,'O5 ('+str(prtl[0])+'%)', sp_dists[0]/max(Order5_nopair[1].apply_map(abs))),
    (Simplex_set,'Simp/O3 ('+str(prtl[0])+'%)', sp_dists[0]/max(Simplex_set[1].apply_map(abs))),
    (Order7_nopair,'O7 ('+str(prtl[1])+'%)', sp_dists[1]/max(Order7_nopair[1].apply_map(abs))),
    (Order5_nopair,'O5 ('+str(prtl[1])+'%)', sp_dists[1]/max(Order5_nopair[1].apply_map(abs))),
    (Simplex_set,'Simp/O3 ('+str(prtl[1])+'%)', sp_dists[1]/max(Simplex_set[1].apply_map(abs)))
]

spoint_sets = [
    (Order7_nopair,'O7', 1),
    (Order5_nopair,'O5', 1),
    (Simplex_set,'Simp/O3', 1),
    (Order7_nopair,'O7 (0.5)', 0.5),
    (Order5_nopair,'O5 (0.5)', 0.5),
    (Simplex_set,'Simp/O3 (0.5)', 0.5)
]

#base_set = (Order7_nopair,'O7')
base_set = (Order5_nopair,'O5')
spoint_scalings = [
    (base_set[0],base_set[1]+' (0.1)',0.1),
    (base_set[0],base_set[1]+' (0.25)',0.25),
    (base_set[0],base_set[1]+' (0.4)',0.4),
    (base_set[0],base_set[1]+' (0.5)',0.5),
    (base_set[0],base_set[1]+' (0.6)',0.6),
    (base_set[0],base_set[1]+' (0.75)',0.75),
    (base_set[0],base_set[1]+' (0.9)',0.9),
    (base_set[0],base_set[1],1)
]
spoint_scalings.reverse()

base_set = (Simplex_set,'Simp/O3')
spoint_scalings2 = [
    (base_set[0],base_set[1]+' (0.1)',0.1),
    (base_set[0],base_set[1]+' (0.25)',0.25),
    (base_set[0],base_set[1]+' (0.4)',0.4),
    (base_set[0],base_set[1]+' (0.5)',0.5),
    (base_set[0],base_set[1]+' (0.6)',0.6),
    (base_set[0],base_set[1]+' (0.75)',0.75),
    (base_set[0],base_set[1]+' (0.9)',0.9),
    (base_set[0],base_set[1],1)
]
spoint_scalings2.reverse()

def run_linearized(s0h,P0,hs,s0) :
    He = matrix( [[diff(hs([x0])[0],x0)]] ).subs({x0:s0[0]})
    ym = hs(s0)+He*(s0h-s0)
    pzz = He*P0*He.transpose()
    pxz = P0*He.transpose()
    return (ym,pzz,pxz)
    
def run_ekf(s0h,P0,hs) :
    He = matrix( [[diff(hs(vector([x0]))[0],x0)]] ).subs({x0:s0h[0]})
    ym = hs(s0h)
    pzz = He*P0*He.transpose()
    pxz = P0*He.transpose()
    return (ym,pzz,pxz)

def run_spoint(s0h,P0_root,hs,spoints,alpha) :
    weights = spoints[0]
    spoints_t = spoints.submatrix(1,0).transpose()
    y0 = hs(s0h)
    ty = [ hs(s0h+alpha*P0_root*spoints_t[j]) for j in range(0,len(weights))]
    ym = sum([weights[j]*ty[j] for j in range(0,len(weights))])
    pzz = sum([ weights[j]*(ty[j]-ym).outer_product(ty[j]-ym) for j in range(0,len(weights))])
    pxz = sum([ weights[j]*(alpha*P0_root*spoints_t[j]).outer_product(ty[j]-ym) for j in range(0,len(weights)) ])
    H =  ((P0_root*P0_root.transpose()).inverse()*pxz).transpose()/alpha^2
    nl = pzz - H*P0_root*P0_root.transpose()*H.transpose()*alpha^2
    pzz = H*P0_root*P0_root.transpose()*H.transpose()+nl/(alpha*alpha)
    pxz = P0_root*P0_root.transpose()*H
    return (ym,pzz,pxz)

def run_spoint_orig(s0h,P0_root,hs,spoints,alpha) :
    weights = spoints[0]
    spoints_t = spoints.submatrix(1,0).transpose()
    y0 = hs(s0h)
    ty = [ (hs(s0h+alpha*P0_root*spoints_t[j])-y0)/alpha^2+y0 for j in range(0,len(weights))]
    ym = sum([weights[j]*ty[j] for j in range(0,len(weights))])
    pzz = sum([ weights[j]*(ty[j]-ym).outer_product(ty[j]-ym) for j in range(0,len(weights))])+2*(y0-ym).outer_product(y0-ym)
    pxz = sum([ weights[j]*(alpha*P0_root*spoints_t[j]).outer_product(ty[j]-ym) for j in range(0,len(weights)) ])
    return (ym,pzz*alpha^2,pxz*alpha)

n_samps = 1000
def run_ensemble(s0h,P0_root,hs) :
    ts = [s0h + P0_root*vector([numpy.random.normal(0,1)]).n() for _ in range(0,n_samps)]
    ty = [hs(ts[j]) for j in range(0,n_samps)]
    ym = vector(numpy.mean(ty,0))
    pyy = matrix(numpy.cov(ts,ty,rowvar=0))
    pzz = pyy.submatrix(1,1,1,1)
    pxz = pyy.submatrix(0,1,1,1)
    return (ym,pzz,pxz)

def check_error(rlz,ym,pzz,pxz) :
    K = pxz*(pzz+R).inverse()
    err = [0 for _ in range(0,len(rlz))]
    for k in range(0,len(rlz)) :
        t_rlz = rlz[k]
        z = h(t_rlz[0],t_rlz[1])
        su = s0h+K*(z-ym)
        err[k] = abs(t_rlz[0][0]-su[0])
    return err

def dump_data(f,data,lbl,lim,tick,k,pre) :
    ndata = len(lbl)
    f.write("summerydata"+str(k)+" = list(\n")
    f.write("\tstats = matrix(\n")
    f.write("\tc"+str(tuple(data))+"\n")
    f.write("\t,5,"+str(ndata)+"),\n")
    f.write("\tn = rep("+str(npoints)+","+str(ndata)+"),\n")
    f.write("\tnames = c"+str(tuple(lbl))+")\n")
    f.write("\n")
    f.write("svg(\'Plots/"+pre+str(k)+".svg',width=6,height="+str(0.2*ndata)+",pointsize=7)\n")
    f.write("par(mar=c(2,7,.1,.1))\n")
    f.write("bxp(summerydata"+str(k)+",show.names='True',horizontal=1,las=1,pars=list(boxcol='white',medlty='blank',medpch=16,medcex=1.0,whisklty=c(1,1),staplelty='blank',outcex=0.5,ylim=c"+str(lim)+",xaxt='n'),axes=F)\n")
    f.write("xtick = c"+str(tuple(tick))+"\n")
    f.write("abline(v=xtick,lty=1,lwd=0.5,col='lightgray')\n")
    f.write("axis(side=1,at=xtick)\n")
    f.write("axis(side=2,c(1:"+str(ndata)+"),summerydata"+str(k)+"$names,las=1)\n")
    f.write("dev.off()\n")
    f.write("\n\n")

scales = [0.5,2,8]
ranges = [
    (0,1),
    (0,4),
    (0,16) 
]
    
ticks = [
    [0,0.25,1],
    [0,0.25,1,2,4],
    [0,0.25,1,2,4,8,16]
]

#Just show that the different methods have trouble
f = open("RegionTrouble.R","w")
err_data = [ [] for _ in range(0,len(spoint_sets_unscaled)+3)]
err_labels = [ "A" for _ in range(0,len(spoint_sets_unscaled)+3)]
for scalek in range(0,len(scales)) :

    scale = scales[scalek]
    P0_root = scale*identity_matrix(1)
    P0 = P0_root*P0_root.transpose()
    print("Standard Deviation : " + str(P0_root[0,0].n(digits=3)))
    print("Variance : " + str(P0[0,0].n(digits=3)))
    hev = h([scale*x0],[0])
    mean = apply_moments(hev)
    print("Truth")
    print("  Expected Value : "+str(mean.n(digits=3)))
    rvar = apply_moments((hev-mean).outer_product(hev-mean).apply_map(expand) )
    print("  Variance : " + str(rvar.n(digits=3)))
    cov = apply_moments( (hev-mean).outer_product(vector([scale*x0])).apply_map(expand) )
    print("  Covariance : " + str(cov.n(digits=3)))

    print("Truth (Linear)")
    hev = h_1([scale*x0],[0])
    mean = apply_moments(hev)
    print("  Expected Value : "+str(mean.n(digits=3)))
    rvar = apply_moments((hev-mean).outer_product(hev-mean).apply_map(expand) )
    print("  Variance : " + str(rvar.n(digits=3)))
    cov = apply_moments( (hev-mean).outer_product(vector([scale*x0])).apply_map(expand) )
    print("  Covariance : " + str(cov.n(digits=3)))
    print("Truth (2nd)")
    hev = h_2([scale*x0],[0])
    mean = apply_moments(hev)
    print("  Expected Value : "+str(mean.n(digits=3)))
    rvar = apply_moments((hev-mean).outer_product(hev-mean).apply_map(expand) )
    print("  Variance : " + str(rvar.n(digits=3)))
    cov = apply_moments( (hev-mean).outer_product(vector([scale*x0])).apply_map(expand) )
    print("  Covariance : " + str(cov.n(digits=3)))
    print("Truth (3nd)")
    hev = h_3([scale*x0],[0])
    mean = apply_moments(hev)
    print("  Expected Value : "+str(mean.n(digits=3)))
    rvar = apply_moments((hev-mean).outer_product(hev-mean).apply_map(expand) )
    print("  Variance : " + str(rvar.n(digits=3)))
    cov = apply_moments( (hev-mean).outer_product(vector([scale*x0])).apply_map(expand) )
    print("  Covariance : " + str(cov.n(digits=3)))

    print("Truth (2nd Fit)")
    hf_2 = fit_2(scale*n(scipy.special.erfinv(.5)*sqrt(2)))
    hev = hf_2([scale*x0])
    mean = apply_moments(hev)
    print("  Expected Value : "+str(mean.n(digits=3)))
    rvar = apply_moments((hev-mean).outer_product(hev-mean).apply_map(expand) )
    print("  Variance : " + str(rvar.n(digits=3)))
    cov = apply_moments( (hev-mean).outer_product(vector([scale*x0])).apply_map(expand) )
    print("  Covariance : " + str(cov.n(digits=3)))

    hev = h([scale*x0],[0])
    mean = apply_moments(hev)

    realizations = [create_realization(P0_root) for _ in range(0,npoints)]

    (ym,pzz,pxz) = run_ekf(s0h,P0.n(),lambda u : h(u,[0]))
    print("EKF")
    print("  Expected Value : " + str(ym.n(digits=3)))
    print("  Variance : " + str(pzz.n(digits=3)))
    print("  Covariance : " + str(pxz.n(digits=3)))
    err = check_error(realizations,ym,pzz,pxz)
    err_data[len(spoint_sets_unscaled)+1] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
    err_labels[len(spoint_sets_unscaled)+1] = "EKF"
    print("  Mean Error : " + str(numpy.mean(err).n(digits=3)))

    for k in range(0,len(spoint_sets_unscaled)) :
        (ym,pzz,pxz) = run_spoint(s0h,P0_root.n(),lambda u : h(u,[0]),spoint_sets_unscaled[k][0],spoint_sets_unscaled[k][2])
        print(spoint_sets_unscaled[k][1])
        print("  Expected Value : " + str(ym.n(digits=3)))
        print("  Variance : " + str(pzz.n(digits=3)))
        print("  Covariance : " + str(pxz.n(digits=3)))
        err = check_error(realizations,ym.n(),pzz.n(),pxz.n())
        err_data[k+1] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
        err_labels[k+1] = spoint_sets_unscaled[k][1]
        print("  Mean Error : " + str(n(numpy.mean(err)).n(digits=3)))

    (ym,pzz,pxz) = run_ensemble(s0h,P0_root.n(),lambda u : h(u,[0]))
    print("Ensemble")
    print("  Expected Value : " + str(ym.n(digits=3)))
    print("  Variance : " + str(pzz.n(digits=3)))
    print("  Covariance : " + str(pxz.n(digits=3)))
    err = check_error(realizations,ym,pzz,pxz)
    err_data[0] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
    err_labels[0] = "Ensemble"
    print("  Mean Error : " + str(n(numpy.mean(err)).n(digits=3)))

    print("Do Nothing")
    err = map(lambda u: abs(u[0][0]),realizations)
    err_data[len(spoint_sets_unscaled)+2] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
    err_labels[len(spoint_sets_unscaled)+2] = "Nothing"
    print("  Mean Error : " + str(n(numpy.mean(err)).n(digits=3)))
    
    dump_data(f,flatten(err_data),err_labels,ranges[scalek],ticks[scalek],scalek,"RegionTrouble")
    
    print("")
    print("")

f.close()

## Original Scaling Method
f = open("RegionScale_orig.R","w")
err_data = [ [] for _ in range(0,len(spoint_scalings)+3)]
err_labels = [ "A" for _ in range(0,len(spoint_scalings)+3)]
for scalek in range(0,len(scales)) :
    scale = scales[scalek]
    P0_root = scale*identity_matrix(1)
    P0 = P0_root*P0_root.transpose()
    print("Standard Deviation : " + str(P0_root[0,0].n(digits=3)))
    print("Variance : " + str(P0[0,0].n(digits=3)))
    hev = h([scale*x0],[0])
    mean = apply_moments(hev)
    print("Truth")
    print("  Expected Value : "+str(mean.n(digits=3)))
    rvar = apply_moments((hev-mean).outer_product(hev-mean).apply_map(expand) )
    print("  Variance : " + str(rvar.n(digits=3)))
    cov = apply_moments( (hev-mean).outer_product(vector([scale*x0])).apply_map(expand) )
    print("  Covariance : " + str(cov.n(digits=3)))

    realizations = [create_realization(P0_root) for _ in range(0,npoints)]

    (ym,pzz,pxz) = run_ekf(s0h,P0.n(),lambda u : h(u,[0]))
    print("EKF")
    print("  Expected Value : " + str(ym.n(digits=3)))
    print("  Variance : " + str(pzz.n(digits=3)))
    print("  Covariance : " + str(pxz.n(digits=3)))
    err = check_error(realizations,ym,pzz,pxz)
    err_data[len(spoint_scalings)+1] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
    err_labels[len(spoint_scalings)+1] = "EKF"
    print("  Mean Error : " + str(numpy.mean(err).n(digits=3)))

    for k in range(0,len(spoint_scalings)) :
        (ym,pzz,pxz) = run_spoint_orig(s0h,P0_root.n(),lambda u : h(u,[0]),spoint_scalings[k][0],spoint_scalings[k][2])
        print(spoint_scalings[k][1])
        print("  Expected Value : " + str(ym.n(digits=3)))
        print("  Variance : " + str(pzz.n(digits=3)))
        print("  Covariance : " + str(pxz.n(digits=3)))
        err = check_error(realizations,ym.n(),pzz.n(),pxz.n())
        err_data[k+1] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
        err_labels[k+1] = spoint_scalings[k][1]
        print("  Mean Error : " + str(n(numpy.mean(err)).n(digits=3)))

    (ym,pzz,pxz) = run_ensemble(s0h,P0_root.n(),lambda u : h(u,[0]))
    print("Ensemble")
    print("  Expected Value : " + str(ym.n(digits=3)))
    print("  Variance : " + str(pzz.n(digits=3)))
    print("  Covariance : " + str(pxz.n(digits=3)))
    err = check_error(realizations,ym,pzz,pxz)
    err_data[0] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
    err_labels[0] = "Ensemble"
    print("  Mean Error : " + str(n(numpy.mean(err)).n(digits=3)))

    print("Do Nothing")
    err = map(lambda u: abs(u[0][0]),realizations)
    err_data[len(spoint_scalings)+2] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
    err_labels[len(spoint_scalings)+2] = "Nothing"
    print("  Mean Error : " + str(n(numpy.mean(err)).n(digits=3)))
    
    dump_data(f,flatten(err_data),err_labels,ranges[scalek],ticks[scalek],scalek,"RegionScale_orig")
    
    print("")
    print("")

f.close()
## New Scaling Method
f = open("RegionScale_new.R","w")
err_data = [ [] for _ in range(0,len(spoint_scalings)+3)]
err_labels = [ "A" for _ in range(0,len(spoint_scalings)+3)]
for scalek in range(0,len(scales)) :
    scale = scales[scalek]
    P0_root = scale*identity_matrix(1)
    P0 = P0_root*P0_root.transpose()
    print("Standard Deviation : " + str(P0_root[0,0].n(digits=3)))
    print("Variance : " + str(P0[0,0].n(digits=3)))
    hev = h([scale*x0],[0])
    mean = apply_moments(hev)
    print("Truth")
    print("  Expected Value : "+str(mean.n(digits=3)))
    rvar = apply_moments((hev-mean).outer_product(hev-mean).apply_map(expand) )
    print("  Variance : " + str(rvar.n(digits=3)))
    cov = apply_moments( (hev-mean).outer_product(vector([scale*x0])).apply_map(expand) )
    print("  Covariance : " + str(cov.n(digits=3)))

    realizations = [create_realization(P0_root) for _ in range(0,npoints)]

    (ym,pzz,pxz) = run_ekf(s0h,P0.n(),lambda u : h(u,[0]))
    print("EKF")
    print("  Expected Value : " + str(ym.n(digits=3)))
    print("  Variance : " + str(pzz.n(digits=3)))
    print("  Covariance : " + str(pxz.n(digits=3)))
    err = check_error(realizations,ym,pzz,pxz)
    err_data[len(spoint_scalings)+1] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
    err_labels[len(spoint_scalings)+1] = "EKF"
    print("  Mean Error : " + str(numpy.mean(err).n(digits=3)))

    for k in range(0,len(spoint_scalings)) :
        (ym,pzz,pxz) = run_spoint(s0h,P0_root.n(),lambda u : h(u,[0]),spoint_scalings[k][0],spoint_scalings[k][2])
        print(spoint_scalings[k][1])
        print("  Expected Value : " + str(ym.n(digits=3)))
        print("  Variance : " + str(pzz.n(digits=3)))
        print("  Covariance : " + str(pxz.n(digits=3)))
        err = check_error(realizations,ym.n(),pzz.n(),pxz.n())
        err_data[k+1] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
        err_labels[k+1] = spoint_scalings[k][1]
        print("  Mean Error : " + str(n(numpy.mean(err)).n(digits=3)))

    (ym,pzz,pxz) = run_ensemble(s0h,P0_root.n(),lambda u : h(u,[0]))
    print("Ensemble")
    print("  Expected Value : " + str(ym.n(digits=3)))
    print("  Variance : " + str(pzz.n(digits=3)))
    print("  Covariance : " + str(pxz.n(digits=3)))
    err = check_error(realizations,ym,pzz,pxz)
    err_data[0] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
    err_labels[0] = "Ensemble"
    print("  Mean Error : " + str(n(numpy.mean(err)).n(digits=3)))

    print("Do Nothing")
    err = map(lambda u: abs(u[0][0]),realizations)
    err_data[len(spoint_scalings)+2] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
    err_labels[len(spoint_scalings)+2] = "Nothing"
    print("  Mean Error : " + str(n(numpy.mean(err)).n(digits=3)))
    
    dump_data(f,flatten(err_data),err_labels,ranges[scalek],ticks[scalek],scalek,"RegionScale_new")
    
    print("")
    print("")

f.close()

##
f = open("RegionScale_new_simp.R","w")
err_data = [ [] for _ in range(0,len(spoint_scalings)+3)]
err_labels = [ "A" for _ in range(0,len(spoint_scalings)+3)]
for scalek in range(0,len(scales)) :
    scale = scales[scalek]
    P0_root = scale*identity_matrix(1)
    P0 = P0_root*P0_root.transpose()
    print("Standard Deviation : " + str(P0_root[0,0].n(digits=3)))
    print("Variance : " + str(P0[0,0].n(digits=3)))
    hev = h([scale*x0],[0])
    mean = apply_moments(hev)
    print("Truth")
    print("  Expected Value : "+str(mean.n(digits=3)))
    rvar = apply_moments((hev-mean).outer_product(hev-mean).apply_map(expand) )
    print("  Variance : " + str(rvar.n(digits=3)))
    cov = apply_moments( (hev-mean).outer_product(vector([scale*x0])).apply_map(expand) )
    print("  Covariance : " + str(cov.n(digits=3)))

    realizations = [create_realization(P0_root) for _ in range(0,npoints)]

    (ym,pzz,pxz) = run_ekf(s0h,P0.n(),lambda u : h(u,[0]))
    print("EKF")
    print("  Expected Value : " + str(ym.n(digits=3)))
    print("  Variance : " + str(pzz.n(digits=3)))
    print("  Covariance : " + str(pxz.n(digits=3)))
    err = check_error(realizations,ym,pzz,pxz)
    err_data[len(spoint_scalings)+1] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
    err_labels[len(spoint_scalings)+1] = "EKF"
    print("  Mean Error : " + str(numpy.mean(err).n(digits=3)))

    for k in range(0,len(spoint_scalings2)) :
        (ym,pzz,pxz) = run_spoint(s0h,P0_root.n(),lambda u : h(u,[0]),spoint_scalings2[k][0],spoint_scalings2[k][2])
        print(spoint_scalings[k][1])
        print("  Expected Value : " + str(ym.n(digits=3)))
        print("  Variance : " + str(pzz.n(digits=3)))
        print("  Covariance : " + str(pxz.n(digits=3)))
        err = check_error(realizations,ym.n(),pzz.n(),pxz.n())
        err_data[k+1] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
        err_labels[k+1] = spoint_scalings2[k][1]
        print("  Mean Error : " + str(n(numpy.mean(err)).n(digits=3)))

    (ym,pzz,pxz) = run_ensemble(s0h,P0_root.n(),lambda u : h(u,[0]))
    print("Ensemble")
    print("  Expected Value : " + str(ym.n(digits=3)))
    print("  Variance : " + str(pzz.n(digits=3)))
    print("  Covariance : " + str(pxz.n(digits=3)))
    err = check_error(realizations,ym,pzz,pxz)
    err_data[0] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
    err_labels[0] = "Ensemble"
    print("  Mean Error : " + str(n(numpy.mean(err)).n(digits=3)))

    print("Do Nothing")
    err = map(lambda u: abs(u[0][0]),realizations)
    err_data[len(spoint_scalings)+2] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
    err_labels[len(spoint_scalings)+2] = "Nothing"
    print("  Mean Error : " + str(n(numpy.mean(err)).n(digits=3)))
    
    dump_data(f,flatten(err_data),err_labels,ranges[scalek],ticks[scalek],scalek,"RegionScale_new_simp")
    
    print("")
    print("")

f.close()

#Try a different scaleing method
err_data = [ [] for _ in range(0,len(spoint_sets)+4)]
err_labels = [ "A" for _ in range(0,len(spoint_sets)+4)]

f = open("RegionExample.R","w")
for scalek in range(0,len(scales)) :
    scale = scales[scalek]
    P0_root = scale*identity_matrix(1)
    P0 = P0_root*P0_root.transpose()
    print("Standard Deviation : " + str(P0_root[0,0].n(digits=3)))
    print("Variance : " + str(P0[0,0].n(digits=3)))
    hev = h([scale*x0],[0])
    mean = apply_moments(hev)
    print("Truth")
    print("  Expected Value : "+str(mean.n(digits=3)))
    rvar = apply_moments((hev-mean).outer_product(hev-mean).apply_map(expand) )
    print("  Variance : " + str(rvar.n(digits=3)))
    cov = apply_moments( (hev-mean).outer_product(vector([scale*x0])).apply_map(expand) )
    print("  Covariance : " + str(cov.n(digits=3)))

    print("Truth (Linear)")
    hev = h_1([scale*x0],[0])
    mean = apply_moments(hev)
    print("  Expected Value : "+str(mean.n(digits=3)))
    rvar = apply_moments((hev-mean).outer_product(hev-mean).apply_map(expand) )
    print("  Variance : " + str(rvar.n(digits=3)))
    cov = apply_moments( (hev-mean).outer_product(vector([scale*x0])).apply_map(expand) )
    print("  Covariance : " + str(cov.n(digits=3)))
    print("Truth (2nd)")
    hev = h_2([scale*x0],[0])
    mean = apply_moments(hev)
    print("  Expected Value : "+str(mean.n(digits=3)))
    rvar = apply_moments((hev-mean).outer_product(hev-mean).apply_map(expand) )
    print("  Variance : " + str(rvar.n(digits=3)))
    cov = apply_moments( (hev-mean).outer_product(vector([scale*x0])).apply_map(expand) )
    print("  Covariance : " + str(cov.n(digits=3)))
    print("Truth (3nd)")
    hev = h_3([scale*x0],[0])
    mean = apply_moments(hev)
    print("  Expected Value : "+str(mean.n(digits=3)))
    rvar = apply_moments((hev-mean).outer_product(hev-mean).apply_map(expand) )
    print("  Variance : " + str(rvar.n(digits=3)))
    cov = apply_moments( (hev-mean).outer_product(vector([scale*x0])).apply_map(expand) )
    print("  Covariance : " + str(cov.n(digits=3)))

    hev = h([scale*x0],[0])
    mean = apply_moments(hev)


    realizations = [create_realization(P0_root) for _ in range(0,npoints)]

    (ym,pzz,pxz) = run_ekf(s0h,P0.n(),lambda u : h(u,[0]))
    print("EKF")
    print("  Expected Value : " + str(ym.n(digits=3)))
    print("  Variance : " + str(pzz.n(digits=3)))
    print("  Covariance : " + str(pxz.n(digits=3)))
    err = check_error(realizations,ym,pzz,pxz)
    err_data[len(spoint_sets)+2] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
    err_labels[len(spoint_sets)+2] = "EKF"
    print("  Mean Error : " + str(numpy.mean(err).n(digits=3)))

    for k in range(0,len(spoint_sets)):
        (ym,pzz,pxz) = run_spoint(s0h,P0_root.n(),lambda u : h(u,[0]),spoint_sets[k][0],spoint_sets[k][2])
        print(spoint_sets[k][1])
        print("  Expected Value : " + str(ym.n(digits=3)))
        print("  Variance : " + str(pzz.n(digits=3)))
        print("  Covariance : " + str(pxz.n(digits=3)))
        err = check_error(realizations,ym.n(),pzz.n(),pxz.n())
        err_data[k+2] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
        err_labels[k+2] = spoint_sets[k][1]
        print("  Mean Error : " + str(n(numpy.mean(err)).n(digits=3)))
        
    
    (ym,pzz,pxz) = run_ensemble(s0h,P0_root.n(),lambda u : h(u,[0]))
    print("Ensemble")
    print("  Expected Value : " + str(ym.n(digits=3)))
    print("  Variance : " + str(pzz.n(digits=3)))
    print("  Covariance : " + str(pxz.n(digits=3)))
    err = check_error(realizations,ym,pzz,pxz)
    err_data[1] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
    err_labels[1] = "Ensemble"
    print("  Mean Error : " + str(n(numpy.mean(err)).n(digits=3)))

    print("Do Nothing")
    err = map(lambda u: abs(u[0][0]),realizations)
    err_data[len(spoint_sets)+3] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
    err_labels[len(spoint_sets)+3] = "Nothing"
    print("  Mean Error : " + str(n(numpy.mean(err)).n(digits=3)))

    print("Perfect KF")
    err = [0 for _ in range(0,len(realizations))]
    for j in range(0,len(realizations)) :
        (ym,pzz,pxz) = run_linearized(s0h,P0.n(),lambda u : h(u,[0]),realizations[j][0])
        K = pxz*(pzz+R).inverse()
        t_rlz = realizations[j]
        z = h(t_rlz[0],t_rlz[1])
        su = s0h+K*(z-ym)
        err[j] = abs(t_rlz[0][0]-su[0])
    err_data[0] = map(n,mquantiles(err,prob=[.1,.25,.5,.75,.9]))
    err_labels[0] = "Perf KF"
    print("  Mean Error : " + str(n(numpy.mean(err)).n(digits=3)))
    
    dump_data(f,flatten(err_data),err_labels,ranges[scalek],ticks[scalek],scalek,"RegionScale")
    
    print("")
    print("")

f.close()

#Iteration Model
npoints = 200

scale = scales[len(scales)-1]
P0_root = scale*identity_matrix(1)
P0 = P0_root*P0_root.transpose()
print("Standard Deviation : " + str(P0_root[0,0].n(digits=3)))
print("Variance : " + str(P0[0,0].n(digits=3)))
hev = h([scale*x0],[0])
mean = apply_moments(hev)

realizations = [create_realization(P0_root) for _ in range(0,npoints)]

niters = 10

blank_data = [0 for _ in range(0,len(realizations))]

err_raw = [ copy(blank_data) for _ in range(0, niters*(1+len(spoint_sets))) ]

for rind in range(0,len(realizations)) :
    t_rlz = realizations[rind]
    lp = s0h
    for itr in range(0,niters) :
        (ym,pzz,pxz) = run_linearized(s0h,P0.n(),lambda u : h(u,[0]),lp)
        K = pxz*(pzz+R).inverse()
        z = h(t_rlz[0],t_rlz[1])
        su = s0h+K*(z-ym)
        err_raw[niters-itr-1][rind] = abs(t_rlz[0][0]-su[0])
        lp = su
    for k in range(0,len(spoint_sets)):
        p0_u = P0.n()
        lp = s0h
        for itr in range(0,niters) :
            tmp = linalg.svd(p0_u)
            p0_u_root = matrix(tmp[0])*diagonal_matrix(vector(tmp[1]).apply_map(sqrt))*matrix(tmp[2])

            (ym,pzz,pxz) = run_spoint(lp,p0_u_root,lambda u : h(u,[0]),spoint_sets[k][0],spoint_sets[k][2])
            z = h(t_rlz[0],t_rlz[1])
            H = (p0_u.inverse()*pxz).transpose()
            nl = pzz-H*p0_u*H.transpose()
            K = P0.n()*H*(H*P0*H.transpose()+R+nl).inverse()
            su = lp+K*(z-ym)
            err_raw[niters*(k+2)-itr-1][rind] = abs(t_rlz[0][0]-su[0])
            lp = su
            tf = (identity_matrix(1)-K*H)
            p0_u = tf*P0*tf.transpose()+K*(R+nl)*K.transpose()

f = open("RegionIter.R","w")
err_data = [ [0 for _ in range(0,5)] for k in range(0,niters)]
err_labels = [ "" for _ in range(0, niters)]
for itr in range(0,niters) :
    err_labels[niters-itr-1] = "EKF Iter "+str(itr)
    err_data[niters-itr-1] = map(n,mquantiles(err_raw[niters-itr-1],prob=[.1,.25,.5,.75,.9]))
dump_data(f,flatten(err_data),err_labels,ranges[scalek],ticks[scalek],0,"RegionIter")
for k in range(0,len(spoint_sets)) :
    for itr in range(0,niters) :
        err_labels[niters-itr-1] = spoint_sets[k][1] + " Iter " + str(itr)
        err_data[niters-itr-1] = map(n,mquantiles(err_raw[niters*(k+2)-itr-1],prob=[.1,.25,.5,.75,.9]))
    dump_data(f,flatten(err_data),err_labels,ranges[scalek],ticks[scalek],k+1,"RegionIter")
f.close()
