#Forplay
reset()
#set_random_seed(52836);

import os

import scipy
from scipy import optimize

import numpy
from numpy import linalg

from sage.misc.decorators import infix_operator
@infix_operator('multiply')
def dot(a,b) : return vector([a[i]*b[i] for i in range(0,min(len(a),len(b)))])

#Actual Stuff

p = 6
depth = 2

def E(d,k,j) :
    ee = matrix(QQ,d)
    ee[k,j]=1
    return ee

def cosine_matrix(n,j,k,theta) : return (cos(theta)-1)*(E(n,j,j)+E(n,k,k))+sin(theta)*(E(n,j,k)-E(n,k,j)) + matrix.identity(n)

def to_cmatrix(n,pp,th) : return prod([cosine_matrix(n,pp[k][0],pp[k][1],th[k]) for k in range(0,len(pp))])

base_vars = reduce(lambda u,v : u+" "+v,["theta"+str(k) for k in range(0,p*(p-1)/2)])
var(base_vars)
base_vlist = eval(base_vars.replace(' ',','))

temp_order = [ [j,k] for j in range(0,p) for k in range(j+1,p) ]
temp_order.reverse()

SP = to_cmatrix(p,temp_order,base_vlist)

print("Prototype Form, Dimension : " + str( p*(p-1)/2 ))
#print(SP)

SP_reduced = SP.submatrix(0,0,depth+1,p)
reduced_vars = reduce( lambda u,v : Set(u) + Set(v), map(lambda u : u.variables(),SP_reduced.list()))
print("Reduced Dimension : " + str(len(reduced_vars)))

S_r = SP_reduced.__copy__()
w_root_i = S_r[0].apply_map(lambda u : 1/u)
S_r[0] = S_r[0] *dot* S_r[0]
for k in range(1,depth+1) :
    S_r[k] = S_r[k] *dot* w_root_i

t_cos_var = ["c"+str(k) for k in range(0,len(reduced_vars))]
t_sin_var = ["s"+str(k) for k in range(0,len(reduced_vars))]
var(reduce(lambda u,v : u+" "+v, t_cos_var+t_sin_var))
t_vars = map(eval,t_cos_var+t_sin_var)

t_cos_subs = {cos(reduced_vars[k]):eval(t_cos_var[k]) for k in range(0,len(reduced_vars))}
t_sin_subs = {sin(reduced_vars[k]):eval(t_sin_var[k]) for k in range(0,len(reduced_vars))}

trig_eqs = [ eval(t_cos_var[k])**2+eval(t_sin_var[k])**2-1 for k in range(0,len(reduced_vars))]

SP_r_s = SP_reduced.subs(t_cos_subs).subs(t_sin_subs)
#print(SP_r_s)

non_zeros = SP_r_s[0]
lcmd = lcm(non_zeros)

S_r_s = SP_r_s.__copy__()
w_root_i = S_r_s[0].apply_map(lambda u : 1/u)
S_r_s[0] = S_r_s[0] *dot* S_r_s[0]
for k in range(1,depth+1) :
    S_r_s[k] = S_r_s[k] *dot* w_root_i

#print("Sigma Points")
#print(S_r_s)

#print("Without common multiple")
#print(S_r_s*lcmd)

def from_condition(S,l,c) : return S[0] * reduce(lambda u,v : u *dot* v,map(lambda w : S[w]*l,c[0])) - c[1]*l**len(c[0])

conditions = [
    [[1,1,1],0],
    [[1,1,2],0],
    [[1,2,2],0],
    [[2,2,2],0],
    [[1,1,1,1],3],
    [[1,1,1,2],0],
    [[1,1,2,2],11],
    [[1,2,2,2],0],
    [[2,2,2,2],3]
]

print("Conditions we wish to impose")
print(conditions)


equations = map(lambda u : from_condition(S_r_s,lcmd,u), conditions) + trig_eqs

print("Using Polynomial Methods, with the substitution cos(theta)=c, sin(theta)=s, c^2+s^2=1")

print("Number of Equations : " + str(len(equations)))
print("Number of Variables : " + str(len(t_vars)))

#f = open("test.phc",'w')
#f.write(str(len(equations))+"\n")
#for k in range(0,len(equations)) :
#    f.write(str(equations[k])+";\n")
#f.close()

print("Using the original Trig Methods")
print("Number of Equations : " + str(len(conditions)))
print("Number of Variables : " + str(len(reduced_vars)))
equations_2 = map(lambda u : from_condition(S_r,1,u), conditions) + [0 for _ in range(0,len(reduced_vars)-len(conditions))]

things_to_swap = Set()

things_to_swap = {theta10,theta6,theta3} #One Negative 6
#things_to_swap = {theta10,theta11,theta6,theta3,theta7,theta4} #Two Negative 6

#things_to_swap = {theta15,theta10,theta6} #One Negative 7

imag_scale = [1 for _ in range(0,len(reduced_vars))]
for k in range(0,len(reduced_vars)) :
    if len( things_to_swap.intersection(Set({reduced_vars[k]}))) > 0 :
        imag_scale[k] = I

def create_tsub(v) :
    return {reduced_vars[k] : imag_scale[k]*v[k] for k in range(0,len(reduced_vars))}

def eval_eqs(v) : 
    tsubs = create_tsub(v)
    vals = map(lambda u : real(u.subs(tsubs)).n(),equations_2)
    min_w = min(map(abs,S_r[0].subs(tsubs).n()))
    if min_w < 0.05 : vals[-1] = 1/min_w
    max_w = max(map(abs,S_r[0].subs(tsubs).n()))
    if max_w > 5 : vals[-2] = max_w
    return vals

def eval_S(v) :
    tsubs = create_tsub(v)
    return(S_r.subs(tsubs).n())

#print(eval_S(tpoint).n(digits=3))

best_sol = [0 for _ in range(0,len(reduced_vars))]
best_e = 1000

for k in range (0,50) :
    tpoint = [ 2*(random()-0.5) for _ in range(0,len(reduced_vars))]
    tsol = optimize.leastsq(eval_eqs,tpoint)
    print("Numeric Solution")
    print(tsol)
    e = eval_eqs(tsol[0])
    err = vector(e)*vector(e)
    print("Is it good? " + str(err.n(digits=3)))
    if (err < 1) : print("ITS GOOD!!!!\n YEH\n YEH")
    if (err < best_e) : 
        best_sol = tsol
        best_e = err
    print(map(lambda u : u.n(digits=3),e))
    print("Actual Solution looks like")
    print(eval_S(tsol[0]).n(digits=3))
print("")
print("Best Error :" + str(best_e.n(digits=3)))
print(eval_S(best_sol[0]))

#macaulay2.restart()
#macaulay2.eval("R = QQ"+str(t_vars))
#macaulay2.eval("I = ideal" + str( tuple(equations)))
#macaulay2.eval("M = I")
#for k in range(0,len(non_zeros)) :
#    macaulay2.eval("M = saturate(M,"+str(non_zeros[k])+")")
#macaulay2.eval("J = gens gb(I,DegreeLimit => 4)")
#print(macaulay2.eval("dim I"))
#print("Dimension of the system")
#print(macaulay2.eval("dim M"))
#print("System")
#print(macaulay2.eval("toString(M)"))

#test = macaulay2.eval("toString(M)").replace("ideal(","[").replace(")","]")

#filename = tmp_filename(ext=".axiom.input")
#f=open(filename,'w')
#f.write("tsol:=radicalSolve("+test+","+str(t_vars)+")")
#f.close()
#axiom.read(filename)
#os.unlink(filename)
#axiom.eval(")set fortran ints2floats off")
#axiom.eval(")set fortran fortlength 1024")
#axiom.eval(")set fortran defaulttype INTEGER")

#for k in range(0,32) :
#    a = axiom.eval("(outputAsFortran(0=0),map(outputAsFortran,tsol("+str(k+1)+")))").replace(".0D0","").replace("DSQRT","sqrt").replace("=","==").partition("L")[0]
#    a = map(lambda u : u.strip(),a.split("\n"))
#    a.remove("")
#    a.remove("")
#    a = map(eval,a)
#    unknowns_r = {a[k].lhs():a[k].rhs() for k in range(0,len(a))}
#    print("Solution "+str(k+1))
#    print(S_r_s.subs(unknowns_r))
