reset()

import numpy
from numpy import linalg
numpy.random.seed(22636) #This seed was maliciosly created

import scipy
from scipy.stats.mstats import mquantiles

import matplotlib
matplotlib.use('SVG')
matplotlib.rc('font',family='M+ 2p',weight='light')
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse

from sage.misc.decorators import infix_operator
@infix_operator('multiply')
def dot(a,b) : return vector([a[i]*b[i] for i in range(0,min(len(a),len(b)))])

if (numpy.random.randint(10000) != 9214) : print("Warning: Your random number generator did not generate the same number as mine, results may vary")

var('phi sigma')

eval_point = {phi:2*pi,sigma:1/1000}
dt = .1
Q_root = matrix([[0.01]]).n()
R_root = matrix([[pi/4]]).n()
P0_root = diagonal_matrix([pi,1/2*pi]).n()

num_steps = 20
num_steps = 10

Phi = matrix([
    [0,1],
    [-phi^2,-2*phi*sigma]
]).subs(eval_point).n()

Lambda = matrix([
    [0],
    [phi^2]
]).subs(eval_point).n()

teig = linalg.eig(Phi)

print("Original System")
F = (matrix(teig[1])*diagonal_matrix(vector(teig[0]*dt).apply_map(exp))*matrix(teig[1]).inverse()).apply_map(real_part)
print("F matrix")
print(F.n(digits=3))

L = (F-identity_matrix(2))*Phi.inverse()*Lambda
print("L matrix")
print(L.n(digits=3))

print("Q matrix")
Q = Q_root*Q_root.transpose()
print(Q.n(digits=3))

#Define the observation function
def obs(s,v) : 
    z = s[0] + v[0]
    while (z>pi) : z = z-2*pi
    while (-z>pi) : z = z+2*pi
    return (z)
def recenter(z,z0) :
    tz = z
    while ( (tz-z0) > pi) : tz = tz-2*pi
    while ( (z0-tz) > pi) : tz = tz+2*pi
    return tz.n()
R = R_root*R_root.transpose()

#Define the initial state mean
print("The initial state")
s0h = vector([0,0]).n()
P0 = P0_root*P0_root.transpose()
print("    Mean")
print(s0h.n(digits=3))
print("    Cov")
print(P0.n(digits=3))

def create_realization() : 
    s0 = s0h + P0_root*vector(numpy.random.normal(0,1,2))
    w_noise = [Q_root*vector(numpy.random.normal(0,1,1)) for k in range(0,num_steps)]
    v_noise = [R_root*vector(numpy.random.normal(0,1,1)) for k in range(0,num_steps)]
    truth_state = [s0 for k in range(0,num_steps+1)]
    time = [0 for k in range(0,num_steps+1)]
    for k in range(1,num_steps+1) :
        truth_state[k] = F*truth_state[k-1] + L*w_noise[k-1]
        time[k] = time[k-1]+dt
    z = [ vector([real_part(obs(truth_state[k],v_noise[k-1])).n()]) for k in range (1,num_steps+1)]
    return (s0,w_noise,v_noise,truth_state,time,z)

(s0,w_noise,v_noise,truth_state,time,z) = create_realization()
print("  Truth")
print(s0.n(digits=3))

print("Wk")
print(map(lambda u : u.n(digits=3),w_noise))

var('theta theta_d v_n')
obs_e = obs([theta, theta_d],[v_n])
print("H")
H = matrix([diff(obs_e,theta), diff(obs_e,theta_d)])
print(H)
print("J")
J = matrix([diff(obs_e,v_n)])
print(J)

print("Vk")
print(map(lambda u:u.n(digits=3),v_noise))

truth_plot = [map(lambda u : u[k], truth_state) for k in range(0,2)]
print("Truth theta")
print(map(lambda u:u.n(digits=3),truth_plot[0]))
print("Truth Dtheta")
print(map(lambda u:u.n(digits=3),truth_plot[1]))

print("The observed values")
print(map(lambda u : u.n(digits=3),z))

#WARNING WARNING
#Is it reasonable that this uses z instead of zz
def run_linearized(ss0h,zz,lin_points,v_points) :
    ss = [ss0h for k in range(0,num_steps+1)]
    pp = [P0 for k in range(0,num_steps+1)]
    resid = [vector([0]) for k in range(0,num_steps)]
    for k in range(1,num_steps+1) :
        s = ss[k-1]
        p = pp[k-1]
        sn = F*s
        pn = F*p*F.transpose() + L*Q*L.transpose()
        ev_point = {theta:lin_points[k][0],v_n:v_points[k-1][0]}
        He = H.subs(ev_point).n()
        Je = J.subs(ev_point).n()
        K = pn*He.transpose()*(He*pn*He.transpose()+Je*R*Je.transpose()).inverse()
        ty = vector([ recenter(obs_e.subs(ev_point).n(),zz[k-1][0]).n()]) + He*(sn-lin_points[k]) + Je*(-v_points[k-1])
        tr = zz[k-1] - ty
        resid[k-1] = tr
        su = sn + K*tr
        tt = identity_matrix(2)-K*He
        pu = tt*pn*tt.transpose()+K*Je*R*Je.transpose()*K.transpose()
        ss[k] = su
        pp[k] = pu
    return (ss,pp,resid)

def run_ekf(ss0h,zz) :
    ss = [ss0h for k in range(0,num_steps+1)]
    pp = [P0 for k in range(0,num_steps+1)]
    resid = [vector([0]) for k in range(0,num_steps)]
    for k in range(1,num_steps+1) :
        s = ss[k-1]
        p = pp[k-1]
        sn = F*s
        pn = F*p*F.transpose() + L*Q*L.transpose()
        ev_point = {theta:sn[0],v_n:0}
        He = H.subs(ev_point).n()
        Je = J.subs(ev_point).n()
        K = pn*He.transpose()*(He*pn*He.transpose()+Je*R*Je.transpose()).inverse()
        ty = vector([ recenter(obs_e.subs(ev_point).n(),zz[k-1][0]).n()])
        tr = zz[k-1] - ty
        resid[k-1] = tr
        su = sn + K*tr
        tt = identity_matrix(2)-K*He
        pu = tt*pn*tt.transpose()+K*Je*R*Je.transpose()*K.transpose()
        ss[k] = su
        pp[k] = pu
    return (ss,pp,resid)

load("Generate_Sigma_Points.sage")

Order5_pair = Generate_Order4_pair(3).n()
Order4_pair = Generate_Order4_pair_v1(3).n()
Order4_pair_5 = Generate_Order4_pair_v1(5).n()

clist = ["#ac2b36"]

spoint_sets = [
    (Order5_pair,"UKF_O5f","UKF (O5f)","#00a250"),
    (Order4_pair,"UKF_O4f","UKF (O4f)","#00a9e4"),
    (Order4_pair_5,"UKF_O4f_5","UKF (O4f 5D)","#336699")
]

def run_spoint(ss0h,zz,spoints,alpha) :
    ss = [ss0h for k in range(0,num_steps+1)]
    pp = [P0 for k in range(0,num_steps+1)]
    resid = [vector([0]) for k in range(0,num_steps)]
    weights = spoints[0]
    for k in range(1,num_steps+1) :
        s = ss[k-1]
        p = pp[k-1]
        sn = F*s
        pn = F*p*F.transpose() + L*Q*L.transpose()
        tmp = linalg.svd(pn)
        pn_root = matrix(tmp[0])*diagonal_matrix(vector(tmp[1]).apply_map(sqrt))*matrix(tmp[2])
        ty = [vector([0]) for j in range (0,len(weights))]
        spoints_t = spoints.submatrix(1,0).transpose()
        for j in range(0,len(weights)) :
            tsp = spoints_t[j]
            ts = sn + alpha*pn_root*vector([tsp[0],tsp[1]])
            tv = alpha*R_root*vector([tsp[2]])
            ty[j] = vector([recenter(obs(ts,tv),zz[k-1][0]).n()])
        ym = sum([weights[j]*ty[j] for j in range(0,len(weights))])
        pzz = sum([ weights[j]*(ty[j]-ym).outer_product(ty[j]-ym) for j in range(0,len(weights))])
        pxz = sum([ weights[j]*(alpha*pn_root*vector([spoints_t[j][0],spoints_t[j][1]])).outer_product(ty[j]-ym) for j in range(1,len(weights))])
        pvz = sum([ weights[j]*(alpha*R_root*vector([spoints_t[j][2]])).outer_product(ty[j]-ym) for j in range(1,len(weights))])
        H = (pn.inverse()*pxz).transpose()/alpha^2
        V = (R.inverse()*pvz).transpose()/alpha^2
        nl = pzz-H*pn*H.transpose()*alpha^2-V*R*V.transpose()*alpha^2
        pzz = H*pn*H.transpose()+V*R*V.transpose()+nl
        pxz = pn*H.transpose()
        K = pxz*pzz.inverse()
        resid[k-1] = zz[k-1] - ym
        su = sn + K*resid[k-1]
        pu = pn - K*pzz*K.transpose()
        ss[k] = su
        pp[k] = pu
    return (ss,pp,resid)

def generate_xt(s_fil,name,col,filename,full_label) :
    fig = plt.figure()
    fig.set_figwidth(3)
    fig.set_figheight(2.5)
    axs = fig.add_axes( (0.6/7,0.4/4,1-.8/7,1-.6/4))

    axs.set_xlabel("Time")
    axs.set_ylabel("Theta")

    tru, = axs.plot(time,truth_plot[0],'k-',linewidth=3.5)
    per, = axs.plot(time,map(lambda u : u[0],s_per),'^-',color='#8F00FF',linewidth=.7,markeredgecolor='k',markeredgewidth=0.7,markersize=6)
    fil, = axs.plot(time,map(lambda u : u[0],s_fil),'v--',color=col,linewidth=1.5,markeredgecolor='k',markersize=5,markeredgewidth=0.5)

    if full_label : 
        axs.legend( (tru,per,fil), ("Truth","Perfect (KF)",name), framealpha=0.7,numpoints=1,prop={'family':'M+ 2p','weight':'ultralight','size':8})
    else : 
        axs.legend( [fil], [name], framealpha=0.7,numpoints=1,prop={'family':'M+ 2p','weight':'ultralight','size':8})

    yticks = [0,(pi).n(digits=2),(-pi).n(digits=2)]
    yticks.sort()
    axs.set_yticks(yticks)
    axs.set_yticklabels(map(lambda u : u.n(digits=2),yticks), fontsize=8, family='sans-serif', weight='light')
    axs.set_xticks([])
    #axs.set_xticks(time,minor=True)
    axs.grid(True)

    fig.savefig('Plots/Wrapping2_Time_'+filename+'.svg')
    plt.close(fig)

##Perfect Kalman Filter
(s_per,p_per,resid_per) = run_linearized(s0h,z,truth_state,v_noise)

conv = (map(lambda u : sqrt(u[0,0]),p_per))
print(map(lambda u : u.n(digits=3),conv))

##Extended Kalman Filter
(s_ekf,p_ekf,resid_ekf) = run_ekf(s0h,z)
#generate_xt(s_ekf,"EKF",clist[0],"EKF",True)

##SPoint
#for k in range(0,len(spoint_sets)) :
#    (s_sp,p_sp,resid_sp) = run_spoint(s0h,z,spoint_sets[k][0],1)
#    generate_xt(s_sp,spoint_sets[k][2],spoint_sets[k][3],spoint_sets[k][1],False)

## Observation 1 Part
P1 = F*P0*F.transpose() + L*Q*L.transpose()

tsvd = linalg.svd(P1)
P1_root = matrix(tsvd[0])*diagonal_matrix(vector(tsvd[1]).apply_map(sqrt))*matrix(tsvd[2])

num_points = 1000
random_points = [[P1_root*vector(numpy.random.normal(0,1,2))+F*s0h,R_root*vector(numpy.random.normal(0,1,1))] for k in range(0,num_points)]
random_theta = map(lambda u:u[0][0],random_points)

random_obs = [recenter(obs(random_points[k][0],random_points[k][1]),z[0][0]) for k in range(0,num_points)]

random_mean = numpy.mean(random_obs)
random_xcorr = linalg.eig(numpy.cov([random_theta,random_obs]))

ze = recenter(obs(F*s0h,vector([0])),z[0][0])
ev_point = {theta:(F*s0h)[0],v_n:0}
He = matrix([[1,0],H.subs(ev_point)[0]])
Je = matrix([[0],J.subs(ev_point)[0]])
H_xcorr = linalg.eig( (He*P1*He.transpose()+Je*R*Je.transpose()).n())

def small_sigma(sn,pn_root,spoints,alpha) :
    weights = spoints[0]
    ty = [vector([0,0]) for j in range (0,len(weights))]
    spoints_t = spoints.submatrix(1,0).transpose()
    for j in range(0,len(weights)) :
        tsp = spoints_t[j]
        ts = sn + alpha*pn_root*vector([tsp[0],tsp[1]])
        tv = alpha*R_root*vector([tsp[2]])
        ty[j] = vector([ts[0],recenter(obs(ts,tv),z[0][0]).n()])
    ym = sum([weights[j]*ty[j] for j in range(0,len(weights))])
    pzz = sum([ weights[j]*(ty[j]-ym).outer_product(ty[j]-ym) for j in range(0,len(weights))])
    pxz = sum([ weights[j]*(alpha*pn_root*vector([spoints_t[j][0],spoints_t[j][1]])).outer_product(ty[j]-ym) for j in range(0,len(weights))])
    pvz = sum([ weights[j]*(alpha*R_root*vector([spoints_t[j][2]])).outer_product(ty[j]-ym) for j in range(0,len(weights))])
    H = ((pn_root*pn_root.transpose()).inverse()*pxz).transpose()/alpha^2
    V = (R.inverse()*pvz).transpose()/alpha^2
    nl = pzz - H*pn_root*pn_root.transpose()*H.transpose()*alpha^2-V*R*V.transpose()*alpha^2
    pzz = H*pn_root*pn_root.transpose()*H.transpose()+V*R*V.transpose()+nl
    return (ym,linalg.eig(pzz),ty)

def generate_obs(smean,xcorr,samples,weights,name,col,filename,full_label) :
    fig = plt.figure()
    fig.set_figwidth(3)
    fig.set_figheight(2.5)
    axs = fig.add_axes( (0.6/7,0.4/4,1-.8/7,1-.6/4))

    axs.set_xlabel("Theta",fontsize=12)
    axs.set_ylabel("Observations",fontsize=12)

    rands, = axs.plot( random_theta,random_obs,'.',markersize=4,color='#888888')
    rand,  = axs.plot( [ze],[random_mean],'x',markersize=7,color='#333333')
    truth, = axs.plot( [truth_state[1][0]], [z[0][0]], 'bo',markersize=7,color='#8f00ff')
    fil,   = axs.plot( [smean[0]],[smean[1]],'x',markersize=7,color=col)
    
    p1 = map(lambda u : u[0], filter(lambda u : u[1]>0,zip(samples,weights)))
    p2 = map(lambda u : u[0], filter(lambda u : u[1]<0,zip(samples,weights)))

    fil_p, = axs.plot( map(lambda u:u[0],p1),map(lambda u:u[1],p1),'o',markersize=4,color=col)
    fil_p2, = axs.plot( map(lambda u:u[0],p2),map(lambda u:u[1],p2),'^',markersize=4,color=col)

    e1 = Ellipse(
        xy=[ (F*s0h)[0], random_mean],
        width=4*sqrt(random_xcorr[0][0]),
        height=4*sqrt(random_xcorr[0][1]),
        angle=-180/pi.n()*atan2(random_xcorr[1][0][1],random_xcorr[1][0][0]))
    e1.set_fill(False)
    e1.set_edgecolor('#333333')
    e1.set_ls('dashed')
    e1.set_lw(2)
    axs.add_artist(e1)

#    e2 = Ellipse(
#        xy=smean,
#        width=4*sqrt(xcorr[0][0]),
#        height=4*sqrt(xcorr[0][1]),
#        angle=-180/pi.n()*atan2(xcorr[1][0][1],xcorr[1][0][0]))
#    e2.set_fill(False)
#    e2.set_edgecolor(col)
#    e2.set_ls('solid')
#    e2.set_lw(1.5)
#    e2.set_clip_box(axs.bbox)
#    axs.add_artist(e2)

    if full_label : 
        axs.legend( (truth,rands,fil_p), ("Measured","Random Samples",name), framealpha=0.7,numpoints=1,prop={'family':'M+ 2p','weight':'ultralight','size':8})
    else : 
        axs.legend( [fil_p], [name], framealpha=0.7,numpoints=1,prop={'family':'M+ 2p','weight':'ultralight','size':8})

    yticks = [z[0][0],ze,z[0][0]-pi.n(),z[0][0]+pi.n()]
    yticks.sort()
    axs.set_yticks(yticks)
    axs.set_yticklabels(map(lambda u : u.n(digits=2),yticks), fontsize=8, family='sans-serif', weight='light')
    yticks_minor = [smean[1]]
    axs.set_yticks(yticks_minor,minor=True)
    xticks = [ truth_state[1][0],(F*s0h)[0]]
    xticks.sort()
    axs.set_xticks(xticks)
    xticks_minor = [smean[0]]
    axs.set_xticks(xticks_minor,minor=True)
    axs.set_xticklabels(map(lambda u : u.n(digits=2),xticks), fontsize=8, family='sans-serif', weight='light')
    axs.grid(True,'both')

    fig.savefig("Plots/Wrapping2_Obs"+filename+".svg")
    plt.close(fig)

generate_obs([(F*s0h)[0],ze],H_xcorr,[[(F*s0h)[0],ze]],[1],"EKF",clist[0],"EKF",True)

for k in range(0,len(spoint_sets)) :
    (smean_sp,xcorr_sp,spoints_sp) = small_sigma(F*s0h,P1_root,spoint_sets[k][0],1)
    generate_obs(smean_sp,xcorr_sp,spoints_sp,spoint_sets[k][0][0],spoint_sets[k][2],spoint_sets[k][3],spoint_sets[k][1],False)

sp_scale = [0 for k in range(0,len(spoint_sets))]
for k in range(0,len(spoint_sets)) :
    sp_points = spoint_sets[k][0].submatrix(1,0).transpose()
    sp_m = max([sqrt(sp_points[j]*sp_points[j]) for j in range(0,len(spoint_sets[k][0][0]))])
    print(spoint_sets[k][1] + " Max : " + str(sp_m.n(digits=3)))
    sp_scale[k] = 0.675/sp_m

for k in range(0,len(spoint_sets)) :
    (s_sp,p_sp,resid_sp) = run_spoint(s0h,z,spoint_sets[k][0],sp_scale[k])
    (smean_sp,xcorr_sp,spoints_sp) = small_sigma(F*s0h,P1_root,spoint_sets[k][0],sp_scale[k])
#    generate_xt(s_sp,spoint_sets[k][2],spoint_sets[k][3],spoint_sets[k][1],False)
#    generate_obs(smean_sp,xcorr_sp,spoints_sp,spoint_sets[k][2],spoint_sets[k][3],spoint_sets[k][1]+"_S",False)

n_runs = 200

def find_err(truth,fil) :
    d_err = [truth[k][0] - fil[k][0] for k in range(0,len(fil))]
    return map(lambda u : sqrt(u*u), d_err)

realizations = [(create_realization(),k) for k in range(0,n_runs)]

@parallel
def perform_run( (new_this_run,k) ) :
    print("Run "  +str(k))
    this_run = (new_this_run[3],new_this_run[1],new_this_run[2],new_this_run[4],new_this_run[5])
    (s_per,p_per,resid_per) = run_linearized(s0h,this_run[4],this_run[0],this_run[2])
    terr_per = find_err(this_run[0],s_per)
    (s_ekf,p_ekf,resid_ekf) = run_ekf(s0h,this_run[4])
    terr_ekf = find_err(this_run[0],s_ekf)
    terr_sp1 = [ [] for k in range(0,len(spoint_sets)) ]
    terr_sps = [ [] for k in range(0,len(spoint_sets)) ]
    for k in range(0,len(spoint_sets)) :
        (s_sp,p_sp,resid_sp) = run_spoint(s0h,this_run[4],spoint_sets[k][0],1)
        terr_sp1[k] = find_err(this_run[0],s_sp)
        (s_sp,p_sp,resid_sp) = run_spoint(s0h,this_run[4],spoint_sets[k][0],sp_scale[k])
        terr_sps[k] = find_err(this_run[0],s_sp)
    return (terr_per,terr_ekf,terr_sp1,terr_sps)

runs = list(perform_run([(realizations[j],) for j in range(0,n_runs)]))

err_per = [ runs[k][1][0] for k in range(0,n_runs) ]
err_ekf = [ runs[k][1][1] for k in range(0,n_runs) ]
err_sp1 = [ [runs[k][1][2][j] for k in range(0,n_runs)] for j in range(0,len(spoint_sets)) ]
err_sps = [ [runs[k][1][3][j] for k in range(0,n_runs)] for j in range(0,len(spoint_sets)) ]

def add_plot(q,td,col,axs,ls) :
    tt = map(lambda u : u+td,time)
    qp = axs.plot(tt,list(q),ls,color=col,linewidth=1.7,markeredgecolor='k',markeredgewidth=0.7,markersize=6)
    return(qp)

def draw_quartile(qt,name,ybnd,ytk,use_old) :
    quant_per = mquantiles(err_per,prob=[qt],axis=0)[0]
    quant_ekf = mquantiles(err_ekf,prob=[qt],axis=0)[0]
    if use_old :
        quant_sp = [ mquantiles(err_sp1[k],prob=[qt],axis=0)[0] for k in range(0,len(spoint_sets))]
    else :
        quant_sp = [ mquantiles(err_sps[k],prob=[qt],axis=0)[0] for k in range(0,len(spoint_sets))]

    fig = plt.figure()
    fig.set_figwidth(7)
    fig.set_figheight(5)
    axs = fig.add_axes( (0.6/7,0.4/4,1-.8/7,1-.6/4) )

    axs.set_xlabel("Time")
    axs.set_ylabel("Error")

    per, = add_plot(quant_per,0,'#8F00FF',axs,'v-')
    ekf, = add_plot(quant_ekf,0.0,clist[0],axs,'v-')
    
    leg_entries = [ per for k in range(0,len(spoint_sets)+2)]
    leg_names = ["" for k in range(0,len(spoint_sets)+2)]
    leg_entries[0] = per
    leg_names[0] = "Perfect (KF)"
    leg_entries[1] = ekf
    leg_names[1] = "EKF"

    for k in range(0,len(spoint_sets)) :
        sp, = add_plot(quant_sp[k],0.0,spoint_sets[k][3],axs,'v-')
        leg_entries[k+2] = sp
        leg_names[k+2] = spoint_sets[k][2]
    axs.legend( tuple(leg_entries) , tuple(leg_names),
                framealpha=0.7,numpoints=1,prop={'family':'M+ 2p','weight':'ultralight','size':10})
    axs.set_xticks( map(lambda u : u*dt,[0,5,10]) )
    yticks = ytk #+ [quant_ekf[num_steps],quant_per[num_steps]] + [quant_sp[k][num_steps] for k in range(0,len(spoint_sets))]
    yticks.sort()
    axs.set_yticks( map(lambda u : n(u,digits=3),yticks) )
    axs.set_ylim( ybnd )
    axs.grid(True)
    
    if use_old :
        fig.savefig('Plots/Wrapping2_Error_'+name+'.svg')
    else :
        fig.savefig('Plots/Wrapping2_Error_'+name+'_S.svg')
    plt.close(fig)

draw_quartile(.5,'50th',(0,3),[0,1,2,3],True)
draw_quartile(.75,'75th',(0,4),[0,2,4],True)
draw_quartile(.9,'90th',(0,8),[0,2,4,6,8],True)
draw_quartile(.5,'50th',(0,3),[0,2],False)
draw_quartile(.75,'75th',(0,4),[0,2,4],False)
draw_quartile(.9,'90th',(0,8),[0,2,4],False)
