source('./blindtricycle_functions.R')

num_runs = 300

if(!exists('gotdata') || !gotdata) {
	run1 = list()
	run1$name = 'Sigma'
	run1$filter_names = c( 
		'UKF O3',
		'UKF O3 (0.75)',
		'UKF O3 (0.5)',
		'UKF O3 (0.25)',
		'UKF O5',
		'UKF O5 (0.75)',
		'UKF O5 (0.5)',
		'UKF O5 (0.25)',
		'UKF O7',
		'UKF O7 (0.75)',
		'UKF O7 (0.5)',
		'UKF O7 (0.25)',
		'UKF O9',
		'UKF O9 (0.75)',
		'UKF O9 (0.5)',
		'UKF O9 (0.25)',
		'UKF O5f',
		'UKF O5f (0.75)',
		'UKF O5f (0.5)',
		'UKF O5f (0.25)'
	       	)
	run2 = run1 #100
	run3 = run1 #010
	run4 = run1 #110
	run5 = run1 #020
	run6 = run1 #120
	
	run1$dataset = pull_in_datasets('ShortRuns/Run_000_Run',run1$name,1:num_runs)
	run2$dataset = pull_in_datasets('ShortRuns/Run_100_Run',run2$name,1:num_runs)
	run3$dataset = pull_in_datasets('ShortRuns/Run_010_Run',run3$name,1:num_runs)
	run4$dataset = pull_in_datasets('ShortRuns/Run_110_Run',run4$name,1:num_runs)
	run5$dataset = pull_in_datasets('ShortRuns/Run_020_Run',run5$name,1:num_runs)
	run6$dataset = pull_in_datasets('ShortRuns/Run_120_Run',run6$name,1:num_runs)

	run7 = list()
	run7$name = "Tune"
	run7$filter_names = c(
		'UKF O7 (0.75)',
		'UKF O7 (0.70)',
		'UKF O7 (0.65)',
		'UKF O7 (0.60)',
		'UKF O7 (0.55)',
		'UKF O7 (0.50)',
		'UKF O7 (0.45)',
		'UKF O7 (0.40)',
		'UKF O7 (0.35)',
		'UKF O7 (0.30)',
		'UKF O7 (0.25)',
		'UKF O7 (0.20)',
		'UKF O7 (0.15)',
		'UKF O7 (0.1)'
	)
	run8 = run7
	run7$dataset = pull_in_datasets('ShortRuns/Run_000_Run',run7$name,1:num_runs)
	run8$dataset = pull_in_datasets('ShortRuns/Run_020_Run',run8$name,1:num_runs)

	gotdata = TRUE
}

### Going to plot just run 4 as a demo
num = 1

ttick = c(0,5,10)/2
xtick = c(-20,0,40,80)
ytick = c(-40,0,40,60)


## Small Run 1
things_to_plot = add_to_things_to_plot(run1$dataset[[num]],'truth_state','Truth','black',1)
generate_xy(things_to_plot,'SmallRun1_xy',xtick,ytick,ttick)
add_legend(things_to_plot$display_name,things_to_plot$colors,things_to_plot$linetypes)
dev.off()

## Small Run 2
things_to_plot = add_to_things_to_plot(run2$dataset[[num]],'truth_state','Truth','black',1)
generate_xy(things_to_plot,'SmallRun2_xy',xtick,ytick,ttick)
add_legend(things_to_plot$display_name,things_to_plot$colors,things_to_plot$linetypes)
dev.off()

## Small Run 3
things_to_plot = add_to_things_to_plot(run3$dataset[[num]],'truth_state','Truth','black',1)
generate_xy(things_to_plot,'SmallRun3_xy',xtick,ytick,ttick)
add_legend(things_to_plot$display_name,things_to_plot$colors,things_to_plot$linetypes)
dev.off()

## Small Run 4
things_to_plot = add_to_things_to_plot(run4$dataset[[num]],'truth_state','Truth','black',1)
generate_xy(things_to_plot,'SmallRun4_xy',xtick,ytick,ttick)
add_legend(things_to_plot$display_name,things_to_plot$colors,things_to_plot$linetypes)
dev.off()

## Small Run 5
things_to_plot = add_to_things_to_plot(run5$dataset[[num]],'truth_state','Truth','black',1)
generate_xy(things_to_plot,'SmallRun5_xy',xtick,ytick,ttick)
add_legend(things_to_plot$display_name,things_to_plot$colors,things_to_plot$linetypes)
dev.off()

## Small Run 6
things_to_plot = add_to_things_to_plot(run6$dataset[[num]],'truth_state','Truth','black',1)
generate_xy(things_to_plot,'SmallRun6_xy',xtick,ytick,ttick)
add_legend(things_to_plot$display_name,things_to_plot$colors,things_to_plot$linetypes)
dev.off()

show_summery <- function(filename=c(),ndata=c(),summerydata=list(),ylm=c(),xtick=c()) {
	svg(sprintf('Plots/%s.svg',filename),width=6,height=0.2*ndata,pointsize=7)
	par(mar=c(2,7,.1,.1))
	bxp(summerydata,show.names='True',horizontal=1,las=1,pars=list(boxcol='white',medlty='blank',medpch=16,medcex=1.0,whisklty=c(1,1),staplelty='blank',outcex=0.5,ylim=ylm,xaxt='n'),axes=F)
	abline(v=xtick,lty=1,lwd=0.5,col='lightgray')
	axis(side=1,at=xtick)
	axis(side=2,c(1:ndata),summerydata$names,las=1)
	dev.off()
}

generate_summery <- function(rn = list()) {
	nruns = length(rn$filter_names)
	summery_data = list()
	summery_data$names = c()
	tdt = c()
	for(num in c(1:nruns)) {
		err = c()
		for(run in c(1:num_runs)){
			tdiff = rn$dataset[[run]][[sprintf('filter%02d_forward_1',num-1)]][,10]-rn$dataset[[run]]$truth_state[,10]
			err[run] = sqrt( tdiff[1]*tdiff[1] + tdiff[2]*tdiff[2] )
		}
		summery_data$names[[num]] = rn$filter_names[[num]]
		tdt = append(tdt,quantile(err,probs=c(.1,.25,.5,.75,.9)))
	}
	summery_data$n = rep(100,nruns)
	summery_data$stats = matrix(tdt,5,nruns)
	return(summery_data)
}

tks = c(0,5,10,15,20,40,60)
show_summery('ShortRun_000',length(run1$filter_names),generate_summery(run1),c(0,60),tks)
show_summery('ShortRun_100',length(run2$filter_names),generate_summery(run2),c(0,60),tks)
show_summery('ShortRun_010',length(run3$filter_names),generate_summery(run3),c(0,60),tks)
show_summery('ShortRun_110',length(run4$filter_names),generate_summery(run4),c(0,60),tks)
show_summery('ShortRun_020',length(run5$filter_names),generate_summery(run5),c(0,60),tks)
show_summery('ShortRun_120',length(run6$filter_names),generate_summery(run6),c(0,60),tks)
show_summery('ShortRun_Tune_000',length(run7$filter_names),generate_summery(run7),c(0,60),tks)
show_summery('ShortRun_Tune_020',length(run8$filter_names),generate_summery(run8),c(0,60),tks)
