source('../../R_bin/SDF-reader.R')
source('../../R_bin/colors.R')
require('plotrix')

pull_in_datasets <- function(base_set = c(), run_name = c(), nums = c()) {
	datasets = list()
	for (num in nums) {
		tdataset = read_sdf_file(sprintf('Results/%s%05d_%s.sdf',base_set,num,run_name))
		datasets[[num]]=tdataset
	}
	return(datasets)
}

add_to_things_to_plot <- function( dataset=list(),name = c(), display_name = c(), color = c(), linetype=c() , cthings_to_plot) {
	if (missing('cthings_to_plot')) {
		cthings_to_plot = list()
		cthings_to_plot$time = dataset$time
		cthings_to_plot$values = list()
		cthings_to_plot$display_name = c()
		cthings_to_plot$colors = c()
		cthings_to_plot$linetypes = c()
		cthings_to_plot$name = c()
		cthings_to_plot$things = 0
	}
	
	num = cthings_to_plot$things
	
	cthings_to_plot$values[[num+1]] = dataset[[name]][c(1,2),]
	cthings_to_plot$display_name[num+1] = display_name
	cthings_to_plot$colors[num+1] = color
	cthings_to_plot$linetypes[num+1] = linetype
	cthings_to_plot$name[num+1] = name
	cthings_to_plot$things = num+1
	return(cthings_to_plot)
}


generate_xy <- function(things_to_plot=list(),subname = '', xtick = c(), ytick = c(), ttick = c()) {
	
	nttp = things_to_plot$things

	svg(filename = sprintf('Plots/xy_%s.svg',subname), width = 6, height = 4, pointsize = 7)

	par(mar=c(4.5,4.5,.1,.1),family='M+ 2c regular')
	this_layout = matrix(0,2,3)
	this_layout[1,] = c(1,2,2)
	this_layout[2,] = c(0,3,3)
	layout(this_layout)
	
	plot(c(),c(),
		xlab='X (m)',ylab='Y (m)',xaxt='n',yaxt='n',
		xlim=xtick[c(1,length(xtick))],ylim=ytick[c(1,length(ytick))])
	abline(v=xtick,h=ytick,lty=1,lwd=0.5,col = "lightgray")
	draw.circle(0,-15,7.5,border = 'black')
	draw.circle(2,15,6.7,border = 'black')
	axis(side=1,at=xtick)
	axis(side=2,at=ytick)

	for(ind in 1:nttp) {
		lines(
			things_to_plot$values[[ind]][1,],
			things_to_plot$values[[ind]][2,],
			type='l',
			lwd=1.5,
			col=things_to_plot$colors[ind],
			lty=things_to_plot$linetypes[ind])
	}

	plot(c(),c(),
                xlab='Time (s)',ylab='Y (m)',
                xaxt='n',yaxt='n',
                xlim=ttick[c(1,length(ttick))],ylim=ytick[c(1,length(ytick))])
	abline(v=ttick,h=ytick,lty=1,lwd=0.5,col = "lightgray")
	axis(side = 1, at = ttick )
	axis(side = 2, at = ytick )

	for(ind in 1:nttp) {
		lines(
			things_to_plot$time,
			things_to_plot$values[[ind]][2,],
			type='l',
			lwd=1.5,
			col=things_to_plot$colors[ind],
			lty=things_to_plot$linetypes[ind])
	}
		
	plot(c(),c(),
		xlab='Time (s)',ylab='X (m)',
                xaxt='n',yaxt='n',
                xlim=ttick[c(1,length(ttick))],ylim=xtick[c(1,length(xtick))])
	abline(v=ttick,h=xtick,lty=1,lwd=0.5,col = "lightgray")
	axis(side = 1, at = ttick )
        axis(side = 2, at = xtick )

	for(ind in 1:nttp) {
		lines(
			things_to_plot$time,
			things_to_plot$values[[ind]][1,],
			type='l',
			lwd=1.5,
			col=things_to_plot$colors[ind],
			lty=things_to_plot$linetypes[ind])
	}
}

xy_small_mult <- function(dataset,subname='',things_to_plot,xtick,ytick) {


	nttp = things_to_plot$things

	svg(filename = sprintf('Plots/xy_%s_smallmult.svg',subname), width = 6, height = 6, pointsize = 7)

	par(mar=c(4.5,4.5,.1,.1),family='M+ 2c regular')
	this_layout = matrix(c(
		1,2,3,
		4,5,6,
		7,8,9),3,3,TRUE)
	layout(this_layout)
	
	for (pind in 1:9) {
		plot(c(),c(),
			xlab='X (m)',ylab='Y (m)',xaxt='n',yaxt='n',
			xlim=xtick[c(1,length(xtick))],ylim=ytick[c(1,length(ytick))])
		abline(v=xtick,h=ytick,lty=1,lwd=0.5,col = "lightgray")
		draw.circle(0,-15,7.5,border = 'black')
		draw.circle(2,15,6.7,border = 'black')
		axis(side=1,at=xtick)
		axis(side=2,at=ytick)

		for(ind in 1:nttp) {
			lines(
				dataset[[pind]][[things_to_plot$name[[ind]]]][1,],
				dataset[[pind]][[things_to_plot$name[[ind]]]][2,],
				type='l',
				lwd=1.5,
				col=things_to_plot$colors[ind],
				lty=things_to_plot$linetypes[ind])
		}

	}
}

add_legend <- function(names=c(),colors=c(),linetypes=c(),ind=2,location='topright') {
	par(mfg=c(1,ind))
	
	legend(location,'(x,y)',
		pch=replicate(length(names),-1),
		lwd=replicate(length(names),2),
		legend=names,
		col=colors,
		lty=linetypes,
		seg.len = 4)
}

add_to_things_to_error <- function( dataset=list(),name = c(), display_name = c(), color = c(), linetype=c() , cthings_to_error) {
	if (missing('cthings_to_error')) {
		cthings_to_error = list()
		cthings_to_error$time = dataset[[1]]$time
		cthings_to_error$values = list()
		cthings_to_error$display_name = c()
		cthings_to_error$colors = c()
		cthings_to_error$linetypes = c()
		cthings_to_error$name = c()
		cthings_to_error$things = 0
	}
	
	num = cthings_to_error$things
	
	nds = length(dataset)
	ntimes = length(dataset[[1]]$time)
	t_x_error = array(0,c(ntimes,nds))
	t_y_error = array(0,c(ntimes,nds))
	for (dnum in 1:nds) {
		t_x_error[,dnum] = abs(dataset[[dnum]]$truth_state[1,] - dataset[[dnum]][[name]][1,])^2
		t_y_error[,dnum] = abs(dataset[[dnum]]$truth_state[2,] - dataset[[dnum]][[name]][2,])^2
	}
	t_xy_error = (t_x_error+t_y_error)
	
	mse_x = c()
	mse_y = c()
	mse_xy = c()

	q50_x = c()
	q50_y = c()
	q50_xy = c()

	q75_x = c()
	q75_y = c()
	q75_xy = c()

	q90_x = c()
	q90_y = c()
	q90_xy = c()

	for (tind in 1:ntimes) {
		mse_x[tind] = sqrt(mean(t_x_error[tind,]))
		mse_y[tind] = sqrt(mean(t_y_error[tind,]))
		mse_xy[tind] = sqrt(mean(t_xy_error[tind,]))

		tq = quantile(t_x_error[tind,],c(0.50,0.75,0.9),names=FALSE)
		q50_x[tind] = sqrt(tq)[1]
		q75_x[tind] = sqrt(tq)[2]
		q90_x[tind] = sqrt(tq)[3]
		
		tq = quantile(t_y_error[tind,],c(0.50,0.75,0.9),names=FALSE)
		q50_y[tind] = sqrt(tq)[1]
		q75_y[tind] = sqrt(tq)[2]
		q90_y[tind] = sqrt(tq)[3]

		tq = quantile(t_xy_error[tind,],c(0.50,0.75,0.9),names=FALSE)
		q50_xy[tind] = sqrt(tq)[1]
		q75_xy[tind] = sqrt(tq)[2]
		q90_xy[tind] = sqrt(tq)[3]
	}
	cthings_to_error$mse_x[[num+1]] = mse_x
	cthings_to_error$mse_y[[num+1]] = mse_y
	cthings_to_error$mse_xy[[num+1]] = mse_xy
	cthings_to_error$q50_x[[num+1]] = q50_x
	cthings_to_error$q50_y[[num+1]] = q50_y
	cthings_to_error$q50_xy[[num+1]] = q50_xy
	cthings_to_error$q75_x[[num+1]] = q75_x
	cthings_to_error$q75_y[[num+1]] = q75_y
	cthings_to_error$q75_xy[[num+1]] = q75_xy
	cthings_to_error$q90_x[[num+1]] = q90_x
	cthings_to_error$q90_y[[num+1]] = q90_y
	cthings_to_error$q90_xy[[num+1]] = q90_xy
	
	cthings_to_error$display_name[num+1] = display_name
	cthings_to_error$colors[num+1] = color
	cthings_to_error$linetypes[num+1] = linetype
	cthings_to_error$name[num+1] = name
	cthings_to_error$things = num+1
	return(cthings_to_error)
}


plot_errors <- function(things_to_plot = c(),subname = '', bounds = c(), xytick = c(), ttick = c(), logaxis = ''){
	
	nttp = things_to_plot$things

	svg(filename = sprintf('Plots/error_%s.svg',subname), width = 6, height = 3, pointsize = 7)

	par(mar=c(4.5,4.5,.3,.3),family='M+ 2c regular')
	this_layout = matrix(c(
		1,
		2,
		3),3,1,TRUE)
#	layout(this_layout)


	sub_t = seq(1,length(things_to_plot$time),10)	
		
	plot(c(),c(),
		xlab='Time (s)',ylab='Position Error (m)',
                xaxt='n',yaxt='n',
                xlim=ttick[c(1,length(ttick))],ylim=bounds,log=logaxis)
	abline(v=ttick,h=xytick,lty=1,lwd=0.5,col = "lightgray")
	axis(side = 1, at = ttick)
        axis(side = 2, at = xytick, sprintf('%.1f',xytick))
	
	if(nttp>0) {
	for(ind in 1:nttp) {
		lines(
			things_to_plot$time,
			things_to_plot$mse_xy[[ind]],
			type='l',
			lwd=1,
			col=things_to_plot$colors[ind],
			lty=things_to_plot$linetypes[ind])
		lines(
			things_to_plot$time,
			things_to_plot$q50_xy[[ind]],
			type='l',
			lwd=1,
			col=makeTransparent(things_to_plot$colors[ind],150),
			lty=things_to_plot$linetypes[ind])
		points(
			things_to_plot$time[sub_t],
			things_to_plot$q50_xy[[ind]][sub_t],
			pch=20,
			lwd=1,
			col=things_to_plot$colors[ind],
			bg =things_to_plot$colors[ind])
#		lines(
#			things_to_plot$time,
#			things_to_plot$q75_xy[[ind]],
#			type='l',
#			lwd=1,
#			col=makeTransparent(things_to_plot$colors[ind],alpha=75),
#			lty=things_to_plot$linetypes[ind])
		points(
			things_to_plot$time[sub_t],
			things_to_plot$q75_xy[[ind]][sub_t],
			pch=23,
			lwd=1,
			col=(things_to_plot$colors[ind]),
			bg=things_to_plot$colors[ind])
#		lines(
#			things_to_plot$time,
#			things_to_plot$max_xy[[ind]],
#			type='l',
#			lwd=1,
#			col=makeTransparent(things_to_plot$colors[ind],alpha=75),
#			lty=things_to_plot$linetypes[ind])
		points(
			things_to_plot$time[sub_t],
			things_to_plot$q90_xy[[ind]][sub_t],
			pch=25,
			lwd=1,
			col=(things_to_plot$colors[ind]),
			bg =(things_to_plot$colors[ind]))
	}
	}
}


compare_plot_errors <- function(things_to_plot = c(),subname = '', bounds = c(), xytick = c(), ttick = c(), logaxis = ''){
	
	nttp = things_to_plot$things

	svg(filename = sprintf('Plots/error_%s.svg',subname), width = 6, height = 3, pointsize = 7)

	par(mar=c(4.5,4.5,.3,.3),family='M+ 2c regular')

	sub_t = seq(1,length(things_to_plot$time),10)	
		
	plot(c(),c(),
		xlab='Time (s)',ylab='Position Error (m)',
                xaxt='n',yaxt='n',
                xlim=ttick[c(1,length(ttick))],ylim=bounds,log=logaxis)
	abline(v=ttick,h=xytick,lty=1,lwd=0.5,col = "lightgray")
	axis(side = 1, at = ttick)
        axis(side = 2, at = xytick, sprintf('%.1f',xytick))
	
	if(nttp>0) {
	for(ind in 1:nttp) {
		lines(
			things_to_plot$time,
			things_to_plot$q50_xy[[ind]],
			type='l',
			lwd=1,
			col=makeTransparent(things_to_plot$colors[ind],150),
			lty=things_to_plot$linetypes[ind])
		points(
			things_to_plot$time[sub_t],
			things_to_plot$q50_xy[[ind]][sub_t],
			pch=20,
			lwd=1,
			col=things_to_plot$colors[ind],
			bg =things_to_plot$colors[ind])
		points(
			things_to_plot$time[sub_t],
			things_to_plot$q75_xy[[ind]][sub_t],
			pch=23,
			lwd=1,
			col=(things_to_plot$colors[ind]),
			bg=things_to_plot$colors[ind])
	}
	}
}
