#include "BlindTricycle.h"

#include "linalg.h"
#include "LinearFilters.h"
#include "EKF.h"
#include "SigmaPoint.h"
#include "Smoother.h"
#include "Space.h"

#include <lapacke.h>


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

int *point_index;
int nvisits;

int nfiles = 1;
char **datafiles;

char majortask[100] = "Nothing Major";
char minortask[100] = "Nothing Minor";

// From http://www.rosshemsley.co.uk/2011/02/creating-a-progress-bar-in-c-or-any-other-console-app/
static inline void loadBar(int x, int n, int r, int w) {
	// Only update r times.
    if ( x % (n/r) != 0 ) return;

    // Calculuate the ratio of complete-to-incomplete.
    float ratio = x/(float)n;
    int   c     = ratio * w;
    // Show the percentage complete.
/*    printf("%3d%% [", (int)(ratio*100) );

    // Show the load bar.
    for (int x=0; x<c; x++)
       printf("=");

    for (int x=c; x<w; x++)
       printf(" ");

    printf("]%s:%s\n",majortask,minortask);*/
    // ANSI Control codes to go back to the
    // previous line and clear it.
}

void flattenN(int datalength, MERRY_FORCING *observations, vec<2> *nvals) {
	for(int i=0;i<datalength;i++){
		if(observations[i].ndx == 0) {
			nvals[i][1] = 0;
		} else if (observations[i].ndx == 1) {
			nvals[i][0] = 0;
		} else {
			nvals[i][0] = 0;
			nvals[i][1] = 0;
		}
	}
}

GAUSSIAN<5> wQ;
GAUSSIAN<2> nQ;

int main(int argc, char **argv){
	int datalength;
	NT *time;
	pos dims[2];
	void *tempdata;
	char basename[MAX_FILENAME];

	RUN_SETTINGS settings;
	int dsetnum;

	GAUSSIAN<7> st0;
	vec<7> truth_st0;
	vec<2> *uvals;
	vec<5> *wvals;
	vec<2> *nvals;
	vec<2> *mvals;

	vec<1> *ys;
	MERRY_FORCING *observations;

	TRICYCLE_FORCING *force;
	vec<7> *truth_state;


	if(!read_in_settings(argv[1],settings)) {
		printf("Couldn't handle settings file \"%s\"\n",argv[1]);
		exit(1);
	}

	sprintf(basename,"Data/%s",settings.basename);

	printf("Running %d Datasets\n",settings.nruns);

	nfiles = 1;
	datafiles = (char**)malloc(nfiles*sizeof(char*));
	datafiles[0] = (char*)malloc(MAX_FILENAME*sizeof(char));

	//Loop over the number of runs
	for (dsetnum = 1; dsetnum<(settings.nruns+1); dsetnum++) {
		printf("Reading in data file %05d\n",dsetnum);
		datalength = read_in_scenario(basename,dsetnum,st0,truth_st0,&uvals,&wvals,wQ,&nvals,nQ,&mvals);
		if (datalength == 0) {
			printf("Bad Scenario\n");
			exit(1);
		}

		if (datalength == 0){
			printf("Failed to read in datafile\nAborting\n");
			exit(1);
		}
		printf("Finished reading the file. Lines Read: %d\n",datalength);
		sprintf(datafiles[0],"Results/%sRun%05d_%s.sdf",settings.basename,dsetnum,settings.runname);

		clean_datafiles();

		time = (NT*)malloc((datalength+1)*sizeof(NT));
		time[0] = 0;
		for (int i=1;i<(datalength+1);i++) {
			time[i] = time[i-1] + dt;
		}
		dims[0] = datalength+1;
		add_datafile("time",'f',sizeof(NT),1,dims,(void*)time,0);

		force = (TRICYCLE_FORCING*)malloc(datalength*sizeof(TRICYCLE_FORCING));
		for(int i=0;i<datalength;i++){
			force[i].v = uvals[i][0];
			force[i].gam = uvals[i][1];
		}
		dims[0] = 2;
		dims[1] = datalength;
		add_datafile("forcing",'f',sizeof(NT),2,dims,(void*)uvals,0);

		dims[0] = 5;
		dims[1] = datalength;
		add_datafile("truth_ws",'f',sizeof(NT),2,dims,(void*)wvals,0);

		truth_state = (vec<7>*)malloc((datalength+1)*sizeof(vec<7>));
		truth_state[0] = truth_st0;
		for(int i=0;i<datalength;i++){
			truth_state[i+1] = moveForward(NULL,NULL,truth_state[i],wvals[i],&force[i]);
		}

		dims[0] = 7;
		dims[1] = datalength+1;
		add_datafile("truth_state",'f',sizeof(NT),2,dims,(void*)truth_state,0);

		ys = (vec<1>*)malloc(datalength*sizeof(vec<1>));
		observations = (MERRY_FORCING*)malloc(datalength*sizeof(MERRY_FORCING));
		for (int i=0;i<datalength;i++) {
			if(mvals[i][0] == 1){
				observations[i].ndx = 0;
				observations[i].center = NULL;
				ys[i] = observer1(NULL,NULL,truth_state[i+1],nvals[i][0],&observations[i]);
				observations[i].center = &ys[i];
			} else if (mvals[i][1] == 1) {
				observations[i].ndx = 1;
				observations[i].center = NULL;
				ys[i] = observer2(NULL,NULL,truth_state[i+1],nvals[i][1],&observations[i]);
				observations[i].center = &ys[i];
			} else {
				ys[i] = vec<1>();
				observations[i].ndx = -1;
				observations[i].center = NULL;
			}
		}
		dims[0] = datalength;
		add_datafile("observations",'f',sizeof(NT),1,dims,(void*)ys,0);

		flattenN(datalength,observations,nvals);
		dims[0] = 2;
		dims[1] = datalength;
		add_datafile("truth_ns",'f',sizeof(NT),2,dims,(void*)nvals,0);

		printf("Running %d Filters\n",settings.nfilter);

		for(int i=0;i<settings.nfilter;i++){
			runFilter(i,datalength,st0,truth_state,wvals,nvals,observations,force,settings.filters[i]);
		}
		printf("Done with data file %05d\n",dsetnum);
		free(time);
		free(ys);
		free(force);
		free(observations);
		free(uvals);
		free(wvals);
		free(nvals);
		free(mvals);
	}

	printf("All Done\n");
	exit(1);

}

int read_in_scenario(char *base, int num, GAUSSIAN<7> &st0, vec<7> &truth_st0, vec<2> **uvals, vec<5> **wvals, GAUSSIAN<5> &wdef, vec<2> **nvals, GAUSSIAN<2> &ndef, vec<2> **mvals) {
	char filename[MAX_INPUT];

	vec<7> *tvec7;
	vec<5> *tvec5;
	vec<2> *tvec2;
	int nnum;
	int datalength;

	sprintf(filename,"%s%05d/initial.txt",base,num);
	nnum = read_in_vfile<7>(&tvec7,filename);
	if (nnum != 9) {
		printf("Read in a bad initial file %s%05d/initial.txt\n",base,num);
		printf("%d lines to the file instead of 9\n",nnum);
		if(nnum > 0)
			free(tvec7);
		return 0;
	}
	truth_st0 = vec<7>(tvec7[0]);
	st0.mean = vec<7>(tvec7[1]);
	st0.cov = mat<7,7>(&tvec7[2]);
	free(tvec7);

	sprintf(filename,"%s%05d/uvals.txt",base,num);
	nnum = read_in_vfile<2>(uvals,filename);
	if (nnum == 0){
		printf("Bad uvals file %s%05d/uvals.txt\n",base,num);
		return 0;
	}
	datalength = nnum;

	sprintf(filename,"%s%05d/wvals.txt",base,num);
	nnum = read_in_vfile<5>(wvals,filename);
	if (nnum != datalength){
		printf("Bad number of data points in uvals or wvals %s%05d\n",base,num);
		printf("Expected %d got %d\n",datalength,nnum);
		free(uvals);
		if (nnum>0)
			free(wvals);
		return 0;
	}

	sprintf(filename,"%s%05d/wdefs.txt",base,num);
	nnum = read_in_vfile<5>(&tvec5,filename);
	if (nnum != 6){
		printf("Read in a bad wdef file %s%05d\n",base,num);
		free(uvals);
		free(wvals);
		if(nnum > 0)
			free(tvec5);
		return 0;
	}
	wdef.mean = tvec5[0];
	wdef.cov = mat<5,5>(&tvec5[1]);
	free(tvec5);

	sprintf(filename,"%s%05d/nvals.txt",base,num);
	nnum = read_in_vfile<2>(nvals,filename);
	if ( nnum != datalength){
		printf("Read in a bad nvals file %s%05d\n",base,num);
		printf("Expected %d got %d\n",datalength,nnum);
		free(uvals);
		free(wvals);
		if(nnum > 0)
			free(nvals);
		return 0;
	}

	sprintf(filename,"%s%05d/ndefs.txt",base,num);
	nnum = read_in_vfile<2>(&tvec2,filename);
	if (nnum != 3){
		printf("Read in a bad ndef file %s%05d\n",base,num);
		free(uvals);
		free(wvals);
		free(nvals);
		if (nnum > 0)
			free(tvec2);
		return 0;
	}
	ndef.mean = tvec2[0];
	ndef.cov = mat<2,2>(&tvec2[1]);
	free(tvec2);

	sprintf(filename,"%s%05d/measures.txt",base,num);
	nnum = read_in_vfile<2>(mvals,filename);
	if (nnum != datalength){
		printf("Read in bad measures file %s%05d\n",base,num);
		printf("Expected %d got %d\n",nnum,datalength);
		free(uvals);
		free(wvals);
		free(nvals);
		if(nnum > 0)
			free(mvals);
		return 0;
	}
	return datalength;
}

void runBackEKF(int datalength, GAUSSIAN<7> *state, GAUSSIAN<7> *stateP, ADVANCEMENT<7,5> *advancement, vec<7> *result, vec<5> *ws, MEASUREMENT<7,1,2> *meas, vec<2> *ns) {
	DUALSTATE<7> *runback;
  GAUSSIAN<5> tempw;
	runback = (DUALSTATE<7>*)malloc((datalength+1)*sizeof(DUALSTATE<7>));
	runback[datalength].state = stateP[datalength].mean;
	mat<7,7> cov = state[datalength-1].cov + mat<7,7>(1E-17);
	runback[datalength].dual = inv(cov)*(stateP[datalength].mean-state[datalength].mean);
	runback[datalength].findcov = false;

	for(int i = (datalength); i > 0; i--){
		loadBar(datalength - i,datalength,20,20);
    if(ws != NULL) {
      tempw = wQ;
			runback[i-1] = runBack(runback[i],state[i-1],stateP[i-1],advancement[i-1],&tempw);
      ws[i-1] = tempw.mean;
    } else
			runback[i-1] = runBack(runback[i],state[i-1],stateP[i-1],advancement[i-1],(GAUSSIAN<5>*)NULL);
	}
	result[0] = runback[0].state;
	for (int i=0;i<datalength;i++) {
		result[i+1] = runback[i+1].state;
		if ( ns != NULL && (meas[i].V.val[0] != 0 || meas[i].V.val[1] !=0) )
				ns[i] = findN(runback[i+1],meas[i],nQ);
				//meas[i].y+meas[i].H*(result[i+1]);
	}
	loadBar(20,20,20,20);
	free(runback);
}

void runBackSigma(int datalength, GAUSSIAN<7> *state, GAUSSIAN<7> *stateP, ADVANCEMENT<7,5> *advancement, GAUSSIAN<7> *result, vec<5> *ws, MEASUREMENT<7,1,2> *meas, vec<2> *ns) {
	DUALSTATE<7> *runback;
  GAUSSIAN<5> tempw;
	runback = (DUALSTATE<7>*)malloc((datalength+1)*sizeof(DUALSTATE<7>));
	runback[datalength].state = stateP[datalength].mean;
	mat<7,7> cov = state[datalength-1].cov + mat<7,7>(1E-17);
	runback[datalength].dual = inv(cov)*(stateP[datalength].mean-state[datalength].mean);
	runback[datalength].findcov = true;

	runback[datalength].ctf = (mat<7,7>*)malloc(sizeof(mat<7,7>));
	runback[datalength].cov = (mat<7,7>*)malloc(sizeof(mat<7,7>));
	*runback[datalength].ctf = inv(cov);
	*runback[datalength].cov = stateP[datalength].cov;

	result[datalength].mean = runback[datalength].state;
	result[datalength].cov = *runback[datalength].cov;

	for(int i = (datalength); i > 0; i--){
		loadBar(datalength - i,datalength,20,20);
    if(ws != NULL) {
      tempw = wQ;
			runback[i-1] = runBack(runback[i],state[i-1],stateP[i-1],advancement[i-1],&tempw);
      ws[i-1] = tempw.mean;
    } else
			runback[i-1] = runBack(runback[i],state[i-1],stateP[i-1],advancement[i-1],(GAUSSIAN<5>*)NULL);
		result[i-1].mean = runback[i-1].state;
		result[i-1].cov = *runback[i-1].cov;
		if ( ns != NULL && (meas[i-1].V.val[0] != 0 || meas[i-1].V.val[1] != 0) )
        ns[i-1] = findN(runback[i],meas[i-1],nQ);
	}
	loadBar(20,20,20,20);
	free(runback[datalength-1].cov);
	free(runback[datalength-1].ctf);
	free(runback);
}

void runEKF(int datalength, GAUSSIAN<7> &st0, TRICYCLE_FORCING *force, MERRY_FORCING *observations, GAUSSIAN<7> *state, GAUSSIAN<7> *stateP, ADVANCEMENT<7,5> *advancement, MEASUREMENT<7,1,2> *measures, vec<7>*linstate, vec<5> *linw, vec<2> *linn) {

	ADVPROTO<7,5,TRICYCLE_FORCING> adv;

	state[0] = st0;
	stateP[0] = st0;

	for(int i=0;i<datalength;i++) {
		loadBar(i,datalength,20,20);

		adv.fun = &moveForward;

		adv.noise = wQ;
		if (linw != NULL) {
      adv.noise.mean = linw[i]; //We do not need to re-address the covariance
		}

		adv.forcing = &force[i];

		//EKF
		adv.state = stateP[i];
		if (linstate != NULL) {
      adv.state.mean = linstate[i]; //We do not need to re-address the covariance
		}

		advancement[i] = ekfAdvance(adv);

		state[i+1] = performAdvancement(stateP[i],wQ,advancement[i]);

		if (observations[i].ndx == 0) {
			MEASPROTO<7,1,2,MERRY_FORCING> meas;
			meas.fun = &observer1;
			meas.state = state[i+1];
			if(linstate != NULL) {
        meas.state.mean = linstate[i+1]; //We do not need to re-adress the covariance
			}
			meas.noise = nQ;
			if(linn != NULL){
        meas.noise.mean = linn[i]; //We do not need to re-address the covariance
			}
			meas.forcing = &observations[i];
			measures[i] = ekfMeasurement(meas);
			stateP[i+1] = performMeasurement(state[i+1],nQ,measures[i]);
		} else if (observations[i].ndx == 1) {
			MEASPROTO<7,1,2,MERRY_FORCING> meas;
			meas.fun = &observer2;
			meas.state = state[i+1];
      if(linstate != NULL) {
        meas.state.mean = linstate[i+1]; //We do not need to re-address the covariance
			}
			meas.noise = nQ;
			if(linn != NULL){
        meas.noise.mean = linn[i];
			}
			meas.forcing = &observations[i];
			measures[i] = ekfMeasurement(meas);
			stateP[i+1] = performMeasurement(state[i+1],nQ,measures[i]);
		} else {
			stateP[i+1] = state[i+1];
		}
	}
	loadBar(20,20,20,20);
}

void runBLSF(int datalength, GAUSSIAN<7> &st0, TRICYCLE_FORCING *force, MERRY_FORCING *observations, GAUSSIAN<7> *state, GAUSSIAN<7> *stateP, ADVANCEMENT<7,5> *advancement, MEASUREMENT<7,1,2> *measures, vec<7>*linstate) {

	ADVPROTO<7,5,TRICYCLE_FORCING> adv;

	state[0] = st0;
	stateP[0] = st0;

	for(int i=0;i<datalength;i++) {
		loadBar(i,datalength,20,20);

		adv.fun = &moveForward;

    adv.noise = wQ;

		adv.forcing = &force[i];

		//EKF
		adv.state = stateP[i];
		if (linstate != NULL) {
      adv.state.mean = linstate[i]; // We dont need to re-address the covariance
		}

		advancement[i] = ekfAdvance(adv);

		state[i+1] = performAdvancement(stateP[i],wQ,advancement[i]);

		if (observations[i].ndx == 0) {
			MEASPROTO<7,1,2,MERRY_FORCING> meas;
			meas.fun = &observer1;
			meas.state = state[i+1];
			if(linstate != NULL) {
        meas.state.mean = linstate[i+1]; //We dont need to re-address the covariance
			}
			meas.noise = nQ;
			meas.forcing = &observations[i];
			measures[i] = ekfMeasurement(meas);
			stateP[i+1] = performMeasurement(state[i+1],nQ,measures[i]);
		} else if (observations[i].ndx == 1) {
			MEASPROTO<7,1,2,MERRY_FORCING> meas;
			meas.fun = &observer2;
			meas.state = state[i+1];
			if(linstate != NULL) {
        meas.state.mean = linstate[i+1]; //We dont need to re-address the covariance
			}
			meas.noise = nQ;
			meas.forcing = &observations[i];
			measures[i] = ekfMeasurement(meas);
			stateP[i+1] = performMeasurement(state[i+1],nQ,measures[i]);
		} else {
			stateP[i+1] = state[i+1];
		}
	}
	loadBar(20,20,20,20);
}

void runMerg1(SIGMAPOINT_SETTINGS<7+5> &sigset_prop, NT alpha, int datalength, GAUSSIAN<7> &st0, TRICYCLE_FORCING *force, MERRY_FORCING *observations, GAUSSIAN<7> *state, GAUSSIAN<7> *stateP, ADVANCEMENT<7,5> *advancement, MEASUREMENT<7,1,2> *measures, GAUSSIAN<7>*linstate) {

	ADVPROTO<7,5,TRICYCLE_FORCING> adv;

	state[0] = st0;
	stateP[0] = st0;

	for(int i=0;i<datalength;i++) {
		loadBar(i,datalength,20,20);

		adv.fun = &moveForward;

		adv.noise = wQ;

		adv.forcing = &force[i];

		adv.state = stateP[i];
		if (linstate != NULL) {
      adv.state = linstate[i];
		}

		advancement[i] = sigmaAdvance(sigset_prop,alpha,adv);

		state[i+1] = performAdvancement(stateP[i],wQ,advancement[i]);

		if (observations[i].ndx == 0) {
			MEASPROTO<7,1,2,MERRY_FORCING> meas;
			meas.fun = &observer1;
			meas.state = state[i+1];
			if(linstate != NULL) {
        meas.state = linstate[i+1];
			}
			meas.noise = nQ;
			meas.forcing = &observations[i];
			measures[i] = ekfMeasurement(meas);
			stateP[i+1] = performMeasurement(state[i+1],nQ,measures[i]);
		} else if (observations[i].ndx == 1) {
			MEASPROTO<7,1,2,MERRY_FORCING> meas;
			meas.fun = &observer2;
			meas.state = state[i+1];
			if(linstate != NULL) {
        meas.state = linstate[i+1];
			}
			meas.noise = nQ;
			meas.forcing = &observations[i];
			measures[i] = ekfMeasurement(meas);
			stateP[i+1] = performMeasurement(state[i+1],nQ,measures[i]);
		} else {
			stateP[i+1] = state[i+1];
		}
	}
	loadBar(20,20,20,20);
}

void runMerg2(SIGMAPOINT_SETTINGS<7+2> &sigset_meas, NT alpha, int datalength, GAUSSIAN<7> &st0, TRICYCLE_FORCING *force, MERRY_FORCING *observations, GAUSSIAN<7> *state, GAUSSIAN<7> *stateP, ADVANCEMENT<7,5> *advancement, MEASUREMENT<7,1,2> *measures, GAUSSIAN<7>*linstate) {

	ADVPROTO<7,5,TRICYCLE_FORCING> adv;

	state[0] = st0;
	stateP[0] = st0;

	for(int i=0;i<datalength;i++) {
		loadBar(i,datalength,20,20);

		adv.fun = &moveForward;

		adv.noise = wQ;

		adv.forcing = &force[i];

		adv.state = stateP[i];
		if (linstate != NULL) {
      adv.state = linstate[i];
		}

		advancement[i] = ekfAdvance(adv);

		state[i+1] = performAdvancement(stateP[i],wQ,advancement[i]);

		if (observations[i].ndx == 0) {
			MEASPROTO<7,1,2,MERRY_FORCING> meas;
			meas.fun = &observer1;
			meas.state = state[i+1];
			if(linstate != NULL) {
        meas.state = linstate[i+1];
			}
			meas.noise = nQ;
			meas.forcing = &observations[i];
			measures[i] = sigmaMeasurement(sigset_meas,alpha,meas);
			stateP[i+1] = performMeasurement(state[i+1],nQ,measures[i]);
      //exit(0);
		} else if (observations[i].ndx == 1) {
			MEASPROTO<7,1,2,MERRY_FORCING> meas;
			meas.fun = &observer2;
			meas.state = state[i+1];
			if(linstate != NULL) {
        meas.state = linstate[i+1];
			}
			meas.noise = nQ;
			meas.forcing = &observations[i];
			measures[i] = sigmaMeasurement(sigset_meas,alpha,meas);
			stateP[i+1] = performMeasurement(state[i+1],nQ,measures[i]);
		} else {
			stateP[i+1] = state[i+1];
		}
	}
	loadBar(20,20,20,20);
}

void runSigma(SIGMAPOINT_SETTINGS<7+5> &sigset_prop,SIGMAPOINT_SETTINGS<7+2> &sigset_meas, NT alpha1, NT alpha2,
  int datalength, GAUSSIAN<7> &st0, TRICYCLE_FORCING *force, MERRY_FORCING *observations, GAUSSIAN<7> *state, GAUSSIAN<7> *stateP, ADVANCEMENT<7,5> *advancement, MEASUREMENT<7,1,2> *measures, GAUSSIAN<7>*linstate) {

	ADVPROTO<7,5,TRICYCLE_FORCING> adv;

	state[0] = st0;
	stateP[0] = st0;

	for(int i=0;i<datalength;i++) {
		loadBar(i,datalength,20,20);

		adv.fun = &moveForward;

		adv.noise = wQ;

		adv.forcing = &force[i];

		adv.state = stateP[i];
		if (linstate != NULL) {
      adv.state = linstate[i];
		}

		advancement[i] = sigmaAdvance(sigset_prop,alpha1,adv);
		state[i+1] = performAdvancement(stateP[i],wQ,advancement[i]);

    if (observations[i].ndx == 0) {
      MEASPROTO<7,1,2,MERRY_FORCING> meas;
      meas.fun = &observer1;
      meas.state = state[i+1];
      if(linstate != NULL) {
        meas.state = linstate[i+1];
      }
      meas.noise = nQ;
      meas.forcing = &observations[i];
      measures[i] = sigmaMeasurement(sigset_meas,alpha2,meas);
      stateP[i+1] = performMeasurement(state[i+1],nQ,measures[i]);
    } else if (observations[i].ndx == 1) {
      MEASPROTO<7,1,2,MERRY_FORCING> meas;
      meas.fun = &observer2;
      meas.state = state[i+1];
      if(linstate != NULL) {
        meas.state = linstate[i+1];
      }
      meas.noise = nQ;
      meas.forcing = &observations[i];
      measures[i] = sigmaMeasurement(sigset_meas,alpha2,meas);
      stateP[i+1] = performMeasurement(state[i+1],nQ,measures[i]);
    } else {
      stateP[i+1] = state[i+1];
    }
  }
	loadBar(20,20,20,20);
}

/*
void expandN(int datalength, MERRY_FORCING *observations, vec<2> *nexpand, vec<1> *nback) {
	for(int i=0;i<datalength;i++){
		if(observations[i].ndx == 0) {
			nexpand[i][0] = nback[i][0];
		} else if (observations[i].ndx == 1) {
			nexpand[i][1] = nback[i][0];
		}
	}
}
*/

void filterTRUTH(int datalength,GAUSSIAN<7> &st0,TRICYCLE_FORCING *force,MERRY_FORCING *observations,int nsmooths,vec<7> *truth_state,vec<5> *wvals,vec<2> *nvals,int fnum) {
	void *tempdata;
	pos dims[3];
	GAUSSIAN<7> *state;
	GAUSSIAN<7> *stateP;
	ADVANCEMENT<7,5> *advancement;
	MEASUREMENT<7,1,2> *measures;
	vec<7> *runback;
	vec<5> *wback;
	vec<2> *nback;
	char tname[100];

	printf("Running Filter TRUTH %d\n",nsmooths);

	state = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	stateP = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	advancement = (ADVANCEMENT<7,5>*)malloc(datalength*sizeof(ADVANCEMENT<7,5>));
	measures = (MEASUREMENT<7,1,2>*)malloc(datalength*sizeof(MEASUREMENT<7,1,2>));
	runback = (vec<7>*)malloc((datalength+1)*sizeof(vec<7>));
	wback = (vec<5>*)malloc(datalength*sizeof(vec<5>));
	nback = (vec<2>*)malloc(datalength*sizeof(vec<2>));

	sprintf(majortask,"Filter %d: TRUTH: ",fnum+1);

	strcpy(minortask,"Forward 1");
	runEKF(datalength,
			st0,
			force,
			observations,
			state,
			stateP,
			advancement,
			measures,
			truth_state,
			wvals,
			nvals);
	tempdata = malloc(7*(datalength+1)*sizeof(NT));
	for (int d=0;d<(datalength+1);d++) {
		((vec<7>*)tempdata)[d] = stateP[d].mean;
	}
	dims[0]=7;
	dims[1]=datalength+1;
	sprintf(tname,"filter%02d_forward_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
	free(tempdata);

	dims[0] = 2;
	dims[1] = datalength+1;
	tempdata = malloc((datalength+1)*sizeof(vec<2>));
	for(int d=0;d<(datalength+1);d++){
		(((vec<2>*)tempdata)[d])[0] = stateP[d].cov[0][0];
		(((vec<2>*)tempdata)[d])[1] = stateP[d].cov[1][1];
	}
	sprintf(tname,"filter%02d_pval",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
	free(tempdata);

	strcpy(minortask,"Back 1");
	runBackEKF(datalength,
			state,
			stateP,
			advancement,
			runback,
			wback,
			measures,
			nback);
	sprintf(tname,"filter%02d_back_1",fnum);
	dims[0] = 7;
	dims[1] = datalength+1;
	add_datafile(tname,'f',sizeof(NT),2,dims,(void*)runback,0);
	dims[0] = 5;
	dims[1] = datalength;
	sprintf(tname,"filter%02d_wval_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,(void*)wback,0);
	dims[0] = 2;
	sprintf(tname,"filter%02d_nval_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,(void*)nback,0);

	for(int s=1;s<nsmooths;s++){
		sprintf(minortask,"Forward %d",s+1);
		runEKF(datalength,
				st0,
				force,
				observations,
				state,
				stateP,
				advancement,
				measures,
				runback,
				wback,
				nback);
		tempdata = malloc(7*(datalength+1)*sizeof(NT));
		for (int d=0;d<(datalength+1);d++) {
			((vec<7>*)tempdata)[d] = stateP[d].mean;
		}
		dims[0]=7;
		dims[1]=datalength+1;
		sprintf(tname,"filter%02d_forward_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
		free(tempdata);

		sprintf(minortask,"Back %d",s+1);
		runBackEKF(datalength,
				state,
				stateP,
				advancement,
				runback,
				wback,
				measures,
				nback);
		sprintf(tname,"filter%02d_back_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,(void*)runback,0);
		dims[0] = 5;
		dims[1] = datalength;
		sprintf(tname,"filter%02d_wval_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,(void*)wback,0);
		dims[0] = 2;
		sprintf(tname,"filter%02d_nval_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,(void*)nback,0);
	}
	free(state);
	free(stateP);
	free(advancement);
	free(measures);
	free(runback);
  free(wback);
  free(nback);
}

void filterEKF(int datalength,GAUSSIAN<7> &st0, TRICYCLE_FORCING *force, MERRY_FORCING *observations, int nsmooths, int fnum){
	void *tempdata;
	pos dims[3];
	GAUSSIAN<7> *state;
	GAUSSIAN<7> *stateP;
	ADVANCEMENT<7,5> *advancement;
	MEASUREMENT<7,1,2> *measures;
	vec<7> *runback;
	char tname[100];

	printf("Running Filter EKF %d\n",nsmooths);

	state = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	stateP = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	advancement = (ADVANCEMENT<7,5>*)malloc(datalength*sizeof(ADVANCEMENT<7,5>));
	measures = (MEASUREMENT<7,1,2>*)malloc(datalength*sizeof(MEASUREMENT<7,1,2>));
	runback = (vec<7>*)malloc((datalength+1)*sizeof(vec<7>));

	sprintf(majortask,"Filter %d: EKF: ",fnum+1);

	strcpy(minortask,"Forward 1");
	runEKF(datalength,
			st0,
			force,
			observations,
			state,
			stateP,
			advancement,
			measures,
			NULL,
			NULL,
			NULL);
	tempdata = malloc(7*(datalength+1)*sizeof(NT));
	for (int d=0;d<(datalength+1);d++) {
		((vec<7>*)tempdata)[d] = stateP[d].mean;
	}
	dims[0]=7;
	dims[1]=datalength+1;
	sprintf(tname,"filter%02d_forward_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
	free(tempdata);

	strcpy(minortask,"Back 1");
	runBackEKF(datalength,
			state,
			stateP,
			advancement,
			runback,
			NULL,
			NULL,
			NULL);
	sprintf(tname,"filter%02d_back_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,(void*)runback,0);

	for(int s=1;s<nsmooths;s++){
		sprintf(minortask,"Forward %d",s+1);
		runEKF(datalength,
				st0,
				force,
				observations,
				state,
				stateP,
				advancement,
				measures,
				runback,
				NULL,
				NULL);
		tempdata = malloc(7*(datalength+1)*sizeof(NT));
		for (int d=0;d<(datalength+1);d++) {
			((vec<7>*)tempdata)[d] = stateP[d].mean;
		}
		dims[0]=7;
		dims[1]=datalength+1;
		sprintf(tname,"filter%02d_forward_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
		free(tempdata);

		sprintf(minortask,"Back %d",s+1);
		runBackEKF(datalength,
				state,
				stateP,
				advancement,
				runback,
				NULL,
				NULL,
				NULL);
		sprintf(tname,"filter%02d_back_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,(void*)runback,0);
	}
	free(state);
	free(stateP);
	free(advancement);
	free(measures);
	free(runback);
}

void filterBLSF1(int datalength, GAUSSIAN<7> &st0, TRICYCLE_FORCING *force, MERRY_FORCING *observations, int nsmooths, int fnum){
	void *tempdata;
	pos dims[3];
	GAUSSIAN<7> *state;
	GAUSSIAN<7> *stateP;
	ADVANCEMENT<7,5> *advancement;
	MEASUREMENT<7,1,2> *measures;
	vec<7> *runback;
	char tname[100];


	printf("Running Filter BLSF1 %d\n",nsmooths);

	state = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	stateP = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	advancement = (ADVANCEMENT<7,5>*)malloc(datalength*sizeof(ADVANCEMENT<7,5>));
	measures = (MEASUREMENT<7,1,2>*)malloc(datalength*sizeof(MEASUREMENT<7,1,2>));
	runback = (vec<7>*)malloc((datalength+1)*sizeof(vec<7>));

	sprintf(majortask,"Filter %d: BLSF: ",fnum+1);

	runback[0] = st0.mean;
	for(int i=0;i<datalength;i++) {
		runback[i+1] = moveForward(NULL,NULL,runback[i],wQ.mean,&force[i]);
	}
	dims[0]=7;
	dims[1]=datalength+1;
	sprintf(tname,"filter%02d_back_0",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,(void*)runback,0);

	strcpy(minortask,"Forward 1");
	runBLSF(datalength,
			st0,
			force,
			observations,
			state,
			stateP,
			advancement,
			measures,
			runback);
	tempdata = malloc((datalength+1)*sizeof(vec<7>));
	for (int d=0;d<(datalength+1);d++) {
		((vec<7>*)tempdata)[d] = stateP[d].mean;
	}
	dims[0]=7;
	dims[1]=datalength+1;
	sprintf(tname,"filter%02d_forward_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
	free(tempdata);

	strcpy(minortask,"Back 1");
	runBackEKF(datalength,
			state,
			stateP,
			advancement,
			runback,
			NULL,
			NULL,
			NULL);
	sprintf(tname,"filter%02d_back_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,(void*)runback,0);

	for(int s=1;s<nsmooths;s++){
		sprintf(minortask,"Forward %d",s+1);
		runBLSF(datalength,
				st0,
				force,
				observations,
				state,
				stateP,
				advancement,
				measures,
				runback);
		tempdata = malloc((datalength+1)*sizeof(vec<7>));
		for (int d=0;d<(datalength+1);d++) {
			((vec<7>*)tempdata)[d] = stateP[d].mean;
		}
		dims[0]=7;
		dims[1]=datalength+1;
		sprintf(tname,"filter%02d_forward_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
		free(tempdata);

		sprintf(minortask,"Back %d",s+1);
		runBackEKF(datalength,
				state,
				stateP,
				advancement,
				runback,
				NULL,
				NULL,
				NULL);
		sprintf(tname,"filter%02d_back_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,(void*)runback,0);
	}
	free(state);
	free(stateP);
	free(advancement);
	free(measures);
	free(runback);
}

#define MAX_OBS 94

void blsfUpdate(int datalength,vec<7> *currentpath,TRICYCLE_FORCING *force,MERRY_FORCING *observations,const int nobs, NT relax){
	mat<MAX_OBS,7> obsmats;
	vec<MAX_OBS> y;
	int obsnum;

	if(MAX_OBS < nobs){
		printf("We have a problem with the BLSF filter, in that we have to many observations, %d obs %d max\n",nobs,MAX_OBS);
		return;
	}

	mat<7,7> ctform;

	mat<7,7> ttform;
	vec<1> ty;
	mat<1,7> tobs;

	obsmats = mat<MAX_OBS,7>((NT)0);
	y = vec<MAX_OBS>((NT)0);

	ctform = mat<7,7>((NT)1);
	obsnum = 0;
	for(int i=0;i<datalength;i++) {
		currentpath[i+1] = moveForward(&ttform,NULL,currentpath[i],wQ.mean,&force[i]);
		ctform = ttform * ctform;
		if (observations[i].ndx == 0) {
			ty = observer1(&tobs,(mat<1,2>*)NULL,currentpath[i+1],nQ.mean, &observations[i]);
			y[obsnum] = -ty[0];
			tobs = tobs*ctform;
			for(int j=0;j<7;j++) {
				obsmats[obsnum][j] = tobs[0][j];
			}
			obsnum++;
		} else if (observations[i].ndx == 1) {
			ty = observer2(&tobs,(mat<1,2>*)NULL,currentpath[i+1],nQ.mean, &observations[i]);
			y[obsnum] = -ty[0];
			tobs = tobs*ctform;
			for(int j=0;j<7;j++) {
				obsmats[obsnum][j] = tobs[0][j];
			}
			obsnum++;
		}
	}
	vec<7> tvec;
	tvec = pinv(obsmats)*y;
	currentpath[0] = (currentpath[0]*(1-relax)) + (tvec*relax);
	while(currentpath[0][2]>(2*3.141593)) {
		currentpath[0][2] = currentpath[0][2] - 2*3.14159;
	}
	while(currentpath[0][2]<(-2*3.141593)) {
			currentpath[0][2] = currentpath[0][2] + 2*3.14159;
	}
}

void filterBLSF2(int datalength,GAUSSIAN<7> &st0, TRICYCLE_FORCING *force, MERRY_FORCING *observations, int nsmooths, NT relax, int fnum){
	pos dims[3];

	vec<7> *currentpath;
	int nobs;

	char tname[100];

	printf("Running Filter BLSF2 %d %f\n",nsmooths,relax);

	currentpath = (vec<7>*)malloc((datalength+1)*sizeof(vec<7>));

	sprintf(majortask,"Filter %d: BLSF: ",fnum+1);

	nobs = 0;
	currentpath[0] = st0.mean;
	for(int i=0;i<datalength;i++) {
		currentpath[i+1] = moveForward(NULL,NULL,currentpath[i],wQ.mean,&force[i]);
		if(observations[i].ndx!=-1)
			nobs++;
	}
	dims[0]=7;
	dims[1]=datalength+1;
	sprintf(tname,"filter%02d_back_0",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,(void*)currentpath,0);

	for(int s=0;s<nsmooths;s++) {
		blsfUpdate(datalength,currentpath,force,observations,nobs,relax);

		for(int i=0;i<datalength;i++) {
			currentpath[i+1] = moveForward(NULL,NULL,currentpath[i],wQ.mean,&force[i]);
		}
		dims[0]=7;
		dims[1]=datalength+1;
		sprintf(tname,"filter%02d_back_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,(void*)currentpath,0);
	}


	free(currentpath);
}

void filterMERGE1(int datalength,GAUSSIAN<7> &st0, TRICYCLE_FORCING *force, MERRY_FORCING *observations, int nsmooths, SIGMAPOINT_SETTINGS<7+5> &sigset_prop, NT alpha, int fnum){
	void *tempdata;
	pos dims[3];
	GAUSSIAN<7> *state;
	GAUSSIAN<7> *stateP;
	ADVANCEMENT<7,5> *advancement;
	MEASUREMENT<7,1,2> *measures;
	GAUSSIAN<7> *runback;
	char tname[100];

	printf("Running Filter MERGE1 %d\n",nsmooths);

	state = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	stateP = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	advancement = (ADVANCEMENT<7,5>*)malloc(datalength*sizeof(ADVANCEMENT<7,5>));
	measures = (MEASUREMENT<7,1,2>*)malloc(datalength*sizeof(MEASUREMENT<7,1,2>));
	runback = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));

	sprintf(majortask,"Filter %d: MERGE1: ",fnum+1);

	strcpy(minortask,"Forward 1");
	runMerg1(sigset_prop,
      alpha,
			datalength,
			st0,
			force,
			observations,
			state,
			stateP,
			advancement,
			measures,
			NULL);
	tempdata = malloc(7*(datalength+1)*sizeof(NT));
	for (int d=0;d<(datalength+1);d++) {
		((vec<7>*)tempdata)[d] = stateP[d].mean;
	}
	dims[0]=7;
	dims[1]=datalength+1;
	sprintf(tname,"filter%02d_forward_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
	free(tempdata);

	strcpy(minortask,"Back 1");
	runBackSigma(datalength,
			state,
			stateP,
			advancement,
			runback,
			NULL,
			NULL,
			NULL);
	tempdata = malloc(7*(datalength+1)*sizeof(NT));
	for (int d=0;d<(datalength+1);d++) {
		((vec<7>*)tempdata)[d] = runback[d].mean;
	}
	dims[0]=7;
	dims[1]=datalength+1;
	sprintf(tname,"filter%02d_back_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,(void*)tempdata,0);
	free(tempdata);

	for(int s=1;s<nsmooths;s++){
		sprintf(minortask,"Forward %d",s+1);
		runMerg1(sigset_prop,
        alpha,
				datalength,
				st0,
				force,
				observations,
				state,
				stateP,
				advancement,
				measures,
				runback);
		tempdata = malloc(7*(datalength+1)*sizeof(NT));
		for (int d=0;d<(datalength+1);d++) {
			((vec<7>*)tempdata)[d] = stateP[d].mean;
		}
		dims[0]=7;
		dims[1]=datalength+1;
		sprintf(tname,"filter%02d_forward_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
		free(tempdata);

		sprintf(minortask,"Back %d",s+1);
		runBackSigma(datalength,
				state,
				stateP,
				advancement,
				runback,
				NULL,
				NULL,
				NULL);
		tempdata = malloc(7*(datalength+1)*sizeof(NT));
		for (int d=0;d<(datalength+1);d++) {
			((vec<7>*)tempdata)[d] = runback[d].mean;
		}
		dims[0]=7;
		dims[1]=datalength+1;
		sprintf(tname,"filter%02d_back_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,(void*)tempdata,0);
		free(tempdata);
	}
	free(state);
	free(stateP);
	free(advancement);
	free(measures);
	free(runback);
}

void filterMERGE2(int datalength,GAUSSIAN<7> &st0, TRICYCLE_FORCING *force, MERRY_FORCING *observations, int nsmooths, SIGMAPOINT_SETTINGS<7+2> &sigset_meas, NT alpha, int fnum) {
	void *tempdata;
	pos dims[3];
	GAUSSIAN<7> *state;
	GAUSSIAN<7> *stateP;
	ADVANCEMENT<7,5> *advancement;
	MEASUREMENT<7,1,2> *measures;
	GAUSSIAN<7> *runback;
	char tname[100];

	printf("Running Filter MERGE2 %d\n",nsmooths);

	state = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	stateP = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	advancement = (ADVANCEMENT<7,5>*)malloc(datalength*sizeof(ADVANCEMENT<7,5>));
	measures = (MEASUREMENT<7,1,2>*)malloc(datalength*sizeof(MEASUREMENT<7,1,2>));
	runback = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));

	sprintf(majortask,"Filter %d: MERGE2: ",fnum+1);

	strcpy(minortask,"Forward 1");
	runMerg2(sigset_meas,
      alpha,
			datalength,
			st0,
			force,
			observations,
			state,
			stateP,
			advancement,
			measures,
			NULL);
	tempdata = malloc(7*(datalength+1)*sizeof(NT));
	for (int d=0;d<(datalength+1);d++) {
		((vec<7>*)tempdata)[d] = stateP[d].mean;
	}
	dims[0]=7;
	dims[1]=datalength+1;
	sprintf(tname,"filter%02d_forward_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
	free(tempdata);

	strcpy(minortask,"Back 1");
	runBackSigma(datalength,
			state,
			stateP,
			advancement,
			runback,
			NULL,
			NULL,
			NULL);
	tempdata = malloc(7*(datalength+1)*sizeof(NT));
	for (int d=0;d<(datalength+1);d++) {
		((vec<7>*)tempdata)[d] = runback[d].mean;
	}
	dims[0]=7;
	dims[1]=datalength+1;
	sprintf(tname,"filter%02d_back_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,(void*)tempdata,0);
	free(tempdata);

	for(int s=1;s<nsmooths;s++){
		sprintf(minortask,"Forward %d",s+1);
		runMerg2(sigset_meas,
        alpha,
				datalength,
				st0,
				force,
				observations,
				state,
				stateP,
				advancement,
				measures,
				runback);
		tempdata = malloc(7*(datalength+1)*sizeof(NT));
		for (int d=0;d<(datalength+1);d++) {
			((vec<7>*)tempdata)[d] = stateP[d].mean;
		}
		dims[0]=7;
		dims[1]=datalength+1;
		sprintf(tname,"filter%02d_forward_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
		free(tempdata);

		sprintf(minortask,"Back %d",s+1);
		runBackSigma(datalength,
				state,
				stateP,
				advancement,
				runback,
				NULL,
				NULL,
				NULL);
		tempdata = malloc(7*(datalength+1)*sizeof(NT));
		for (int d=0;d<(datalength+1);d++) {
			((vec<7>*)tempdata)[d] = runback[d].mean;
		}
		dims[0]=7;
		dims[1]=datalength+1;
		sprintf(tname,"filter%02d_back_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,(void*)tempdata,0);
		free(tempdata);
	}
	free(state);
	free(stateP);
	free(advancement);
	free(measures);
	free(runback);
}

void filterSIGMA(int datalength,GAUSSIAN<7> &st0, TRICYCLE_FORCING *force, MERRY_FORCING *observations, int nsmooths,
  SIGMAPOINT_SETTINGS<7+5> &sigset_prop, SIGMAPOINT_SETTINGS<7+2> &sigset_meas, NT alpha1, NT alpha2, int fnum){
	void *tempdata;
	pos dims[3];
	GAUSSIAN<7> *state;
	GAUSSIAN<7> *stateP;
	ADVANCEMENT<7,5> *advancement;
	MEASUREMENT<7,1,2> *measures;
	GAUSSIAN<7> *runback;
	char tname[100];

	printf("Running Filter SIGMA %d\n",nsmooths);

	state = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	stateP = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	advancement = (ADVANCEMENT<7,5>*)malloc(datalength*sizeof(ADVANCEMENT<7,5>));
	measures = (MEASUREMENT<7,1,2>*)malloc(datalength*sizeof(MEASUREMENT<7,1,2>));
	runback = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));

	sprintf(majortask,"Filter %d: SIGMA: ",fnum+1);

	strcpy(minortask,"Forward 1");
	runSigma(sigset_prop,
      sigset_meas,
      alpha1,
      alpha2,
			datalength,
			st0,
			force,
			observations,
			state,
			stateP,
			advancement,
			measures,
			NULL);
	tempdata = malloc(7*(datalength+1)*sizeof(NT));
	for (int d=0;d<(datalength+1);d++) {
		((vec<7>*)tempdata)[d] = stateP[d].mean;
	}
	dims[0]=7;
	dims[1]=datalength+1;
	sprintf(tname,"filter%02d_forward_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
	free(tempdata);

	strcpy(minortask,"Back 1");
	runBackSigma(datalength,
			state,
			stateP,
			advancement,
			runback,
			NULL,
			NULL,
			NULL);
	tempdata = malloc(7*(datalength+1)*sizeof(NT));
	for (int d=0;d<(datalength+1);d++) {
		((vec<7>*)tempdata)[d] = runback[d].mean;
	}
	dims[0]=7;
	dims[1]=datalength+1;
	sprintf(tname,"filter%02d_back_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,(void*)tempdata,0);
	free(tempdata);

	for(int s=1;s<nsmooths;s++){
		sprintf(minortask,"Forward %d",s+1);
		runSigma(sigset_prop,
        sigset_meas,
        alpha1,
        alpha2,
				datalength,
				st0,
				force,
				observations,
				state,
				stateP,
				advancement,
				measures,
				runback);
		tempdata = malloc(7*(datalength+1)*sizeof(NT));
		for (int d=0;d<(datalength+1);d++) {
			((vec<7>*)tempdata)[d] = stateP[d].mean;
		}
		dims[0]=7;
		dims[1]=datalength+1;
		sprintf(tname,"filter%02d_forward_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
		free(tempdata);

		sprintf(minortask,"Back %d",s+1);
		runBackSigma(datalength,
				state,
				stateP,
				advancement,
				runback,
				NULL,
				NULL,
				NULL);
		tempdata = malloc(7*(datalength+1)*sizeof(NT));
		for (int d=0;d<(datalength+1);d++) {
			((vec<7>*)tempdata)[d] = runback[d].mean;
		}
		dims[0]=7;
		dims[1]=datalength+1;
		sprintf(tname,"filter%02d_back_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,(void*)tempdata,0);
		free(tempdata);
	}
	free(state);
	free(stateP);
	free(advancement);
	free(measures);
	free(runback);
}

void filterFP_SIGMA(int datalength,GAUSSIAN<7> &st0, TRICYCLE_FORCING *force, MERRY_FORCING *observations, int nsmooths,
  SIGMAPOINT_SETTINGS<7+5> &sigset_prop, SIGMAPOINT_SETTINGS<7+2> &sigset_meas, NT alpha1, NT alpha2, int fnum){
	void *tempdata;
	pos dims[3];
	GAUSSIAN<7> *state;
	GAUSSIAN<7> *stateP;
	ADVANCEMENT<7,5> *advancement;
	MEASUREMENT<7,1,2> *measures;
	vec<7> *runback;
	char tname[100];

	printf("Running Filter FP_SIGMA %d \n",nsmooths);

	state = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	stateP = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	advancement = (ADVANCEMENT<7,5>*)malloc(datalength*sizeof(ADVANCEMENT<7,5>));
	measures = (MEASUREMENT<7,1,2>*)malloc(datalength*sizeof(MEASUREMENT<7,1,2>));
	runback = (vec<7>*)malloc((datalength+1)*sizeof(vec<7>));

	sprintf(majortask,"Filter %d: SIGMA: ",fnum+1);

	strcpy(minortask,"Forward 1");
	runSigma(sigset_prop,
      sigset_meas,
      alpha1,
      alpha2,
			datalength,
			st0,
			force,
			observations,
			state,
			stateP,
			advancement,
			measures,
			NULL);
	tempdata = malloc(7*(datalength+1)*sizeof(NT));
	for (int d=0;d<(datalength+1);d++) {
		((vec<7>*)tempdata)[d] = stateP[d].mean;
	}
	dims[0]=7;
	dims[1]=datalength+1;
	sprintf(tname,"filter%02d_forward_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
	free(tempdata);

	strcpy(minortask,"Back 1");
	runBackEKF(datalength,
			state,
			stateP,
			advancement,
			runback,
			NULL,
			NULL,
			NULL);
	sprintf(tname,"filter%02d_back_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,(void*)runback,0);

	for(int s=1;s<nsmooths;s++){
		sprintf(minortask,"Forward %d",s+1);
		runEKF(datalength,
				st0,
				force,
				observations,
				state,
				stateP,
				advancement,
				measures,
				runback,
				NULL,
				NULL);
		tempdata = malloc(7*(datalength+1)*sizeof(NT));
		for (int d=0;d<(datalength+1);d++) {
			((vec<7>*)tempdata)[d] = stateP[d].mean;
		}
		dims[0]=7;
		dims[1]=datalength+1;
		sprintf(tname,"filter%02d_forward_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
		free(tempdata);

		sprintf(minortask,"Back %d",s+1);
		runBackEKF(datalength,
				state,
				stateP,
				advancement,
				runback,
				NULL,
				NULL,
				NULL);
		sprintf(tname,"filter%02d_back_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,(void*)runback,0);
	}
	free(state);
	free(stateP);
	free(advancement);
	free(measures);
	free(runback);
}

void filterFP_MERGE1(int datalength,GAUSSIAN<7> &st0, TRICYCLE_FORCING *force, MERRY_FORCING *observations, int nsmooths, SIGMAPOINT_SETTINGS<7+5> &sigset_prop, NT alpha, int fnum){
	void *tempdata;
	pos dims[3];
	GAUSSIAN<7> *state;
	GAUSSIAN<7> *stateP;
	ADVANCEMENT<7,5> *advancement;
	MEASUREMENT<7,1,2> *measures;
	vec<7> *runback;
	char tname[100];

	printf("Running Filter FP_MERGE1 %d\n",nsmooths);

	state = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	stateP = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	advancement = (ADVANCEMENT<7,5>*)malloc(datalength*sizeof(ADVANCEMENT<7,5>));
	measures = (MEASUREMENT<7,1,2>*)malloc(datalength*sizeof(MEASUREMENT<7,1,2>));
	runback = (vec<7>*)malloc((datalength+1)*sizeof(vec<7>));

	sprintf(majortask,"Filter %d: FP_MERGE1: ",fnum+1);

	strcpy(minortask,"Forward 1");
	runMerg1(sigset_prop,
      alpha,
			datalength,
			st0,
			force,
			observations,
			state,
			stateP,
			advancement,
			measures,
			NULL);
	tempdata = malloc(7*(datalength+1)*sizeof(NT));
	for (int d=0;d<(datalength+1);d++) {
		((vec<7>*)tempdata)[d] = stateP[d].mean;
	}
	dims[0]=7;
	dims[1]=datalength+1;
	sprintf(tname,"filter%02d_forward_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
	free(tempdata);

	strcpy(minortask,"Back 1");
	runBackEKF(datalength,
			state,
			stateP,
			advancement,
			runback,
			NULL,
			NULL,
			NULL);
	sprintf(tname,"filter%02d_back_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,(void*)runback,0);

	for(int s=1;s<nsmooths;s++){
		sprintf(minortask,"Forward %d",s+1);
		runEKF(datalength,
				st0,
				force,
				observations,
				state,
				stateP,
				advancement,
				measures,
				runback,
				NULL,
				NULL);
		tempdata = malloc(7*(datalength+1)*sizeof(NT));
		for (int d=0;d<(datalength+1);d++) {
			((vec<7>*)tempdata)[d] = stateP[d].mean;
		}
		dims[0]=7;
		dims[1]=datalength+1;
		sprintf(tname,"filter%02d_forward_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
		free(tempdata);

		sprintf(minortask,"Back %d",s+1);
		runBackEKF(datalength,
				state,
				stateP,
				advancement,
				runback,
				NULL,
				NULL,
				NULL);
		sprintf(tname,"filter%02d_back_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,(void*)runback,0);
	}
	free(state);
	free(stateP);
	free(advancement);
	free(measures);
	free(runback);
}

void filterFP_MERGE2(int datalength,GAUSSIAN<7> &st0, TRICYCLE_FORCING *force, MERRY_FORCING *observations, int nsmooths, SIGMAPOINT_SETTINGS<7+2> &sigset_meas, NT alpha, int fnum){
	void *tempdata;
	pos dims[3];
	GAUSSIAN<7> *state;
	GAUSSIAN<7> *stateP;
	ADVANCEMENT<7,5> *advancement;
	MEASUREMENT<7,1,2> *measures;
	vec<7> *runback;
	char tname[100];

	printf("Running Filter FP_MERGE2 %d\n",nsmooths);

	state = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	stateP = (GAUSSIAN<7>*)malloc((datalength+1)*sizeof(GAUSSIAN<7>));
	advancement = (ADVANCEMENT<7,5>*)malloc(datalength*sizeof(ADVANCEMENT<7,5>));
	measures = (MEASUREMENT<7,1,2>*)malloc(datalength*sizeof(MEASUREMENT<7,1,2>));
	runback = (vec<7>*)malloc((datalength+1)*sizeof(vec<7>));

	sprintf(majortask,"Filter %d: FP_MERGE2: ",fnum+1);

	strcpy(minortask,"Forward 1");
	runMerg2(sigset_meas,
      alpha,
			datalength,
			st0,
			force,
			observations,
			state,
			stateP,
			advancement,
			measures,
			NULL);
	tempdata = malloc(7*(datalength+1)*sizeof(NT));
	for (int d=0;d<(datalength+1);d++) {
		((vec<7>*)tempdata)[d] = stateP[d].mean;
	}
	dims[0]=7;
	dims[1]=datalength+1;
	sprintf(tname,"filter%02d_forward_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
	free(tempdata);

	strcpy(minortask,"Back 1");
	runBackEKF(datalength,
			state,
			stateP,
			advancement,
			runback,
			NULL,
			NULL,
			NULL);
	sprintf(tname,"filter%02d_back_1",fnum);
	add_datafile(tname,'f',sizeof(NT),2,dims,(void*)runback,0);

	for(int s=1;s<nsmooths;s++){
		sprintf(minortask,"Forward %d",s+1);
		runEKF(datalength,
				st0,
				force,
				observations,
				state,
				stateP,
				advancement,
				measures,
				runback,
				NULL,
				NULL);
		tempdata = malloc(7*(datalength+1)*sizeof(NT));
		for (int d=0;d<(datalength+1);d++) {
			((vec<7>*)tempdata)[d] = stateP[d].mean;
		}
		dims[0]=7;
		dims[1]=datalength+1;
		sprintf(tname,"filter%02d_forward_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,tempdata,0);
		free(tempdata);

		sprintf(minortask,"Back %d",s+1);
		runBackEKF(datalength,
				state,
				stateP,
				advancement,
				runback,
				NULL,
				NULL,
				NULL);
		sprintf(tname,"filter%02d_back_%d",fnum,s+1);
		add_datafile(tname,'f',sizeof(NT),2,dims,(void*)runback,0);
	}
	free(state);
	free(stateP);
	free(advancement);
	free(measures);
	free(runback);
}

void runFilter(int fnum, int datalength, GAUSSIAN<7> &st0, vec<7> *truth_state, vec<5> *wvals, vec<2> *nvals, MERRY_FORCING *measures, TRICYCLE_FORCING *force, LINEAR_FILTER &filtersettings){
	GAUSSIAN<7> st0_perfect;

	switch(filtersettings.filter_type){
	case TRUTH_SW:
		filterTRUTH(datalength,st0,force,measures,filtersettings.nsmooths,truth_state,wvals,NULL,fnum);
		break;
	case TRUTH_S_:
		filterTRUTH(datalength,st0,force,measures,filtersettings.nsmooths,truth_state,NULL,NULL,fnum);
		break;
	case TRUTH__W:
		filterTRUTH(datalength,st0,force,measures,filtersettings.nsmooths,NULL,wvals,NULL,fnum);
		break;
	case EKF:
		filterEKF(datalength,st0,force,measures,filtersettings.nsmooths,fnum);
		break;
	case BSEKF:
		filterTRUTH(datalength,st0,force,measures,filtersettings.nsmooths,NULL,NULL,NULL,fnum);
		break;
	case BLSF2:
		filterBLSF2(datalength,st0,force,measures,filtersettings.nsmooths,filtersettings.alpha[0],fnum);
		break;
	case TBLSF2:
		st0_perfect.cov = st0.cov;
		st0_perfect.mean = truth_state[0];
		filterBLSF2(datalength,st0_perfect,force,measures,filtersettings.nsmooths,filtersettings.alpha[0],fnum);
		break;
	case BLSF1:
		filterBLSF1(datalength,st0,force,measures,filtersettings.nsmooths,fnum);
		break;
	case MERGE1:
		filterMERGE1(datalength,st0,force,measures,filtersettings.nsmooths,filtersettings.sp_set_prop,filtersettings.alpha[0],fnum);
		break;
	case MERGE2:
		filterMERGE2(datalength,st0,force,measures,filtersettings.nsmooths,filtersettings.sp_set_meas,filtersettings.alpha[0],fnum);
		break;
	case SIGMA:
		filterSIGMA(datalength,st0,force,measures,filtersettings.nsmooths,
      filtersettings.sp_set_prop,filtersettings.sp_set_meas,filtersettings.alpha[0],filtersettings.alpha[1],fnum);
		break;
	case FP_MERGE1:
		filterFP_MERGE1(datalength,st0,force,measures,filtersettings.nsmooths,filtersettings.sp_set_prop,filtersettings.alpha[0],fnum);
		break;
	case FP_MERGE2:
		filterFP_MERGE2(datalength,st0,force,measures,filtersettings.nsmooths,filtersettings.sp_set_meas,filtersettings.alpha[0],fnum);
		break;
	case FP_SIGMA:
		filterFP_SIGMA(datalength,st0,force,measures,filtersettings.nsmooths,
      filtersettings.sp_set_prop,filtersettings.sp_set_meas,filtersettings.alpha[0],filtersettings.alpha[1],fnum);
		break;
	}
	return;
}

void strip_nl(char *str){
	int len = strlen(str);
	if (str[len-1]=='\n')
		str[len-1] = '\0';
}

int read_in_settings(const char *filename,RUN_SETTINGS &settings) {
	FILE *fp;
	char instring[MAX_INPUT];
	float tempfloat[6];

	printf("Opening settings file %s\n",filename);
	fp = fopen(filename,"r");
	if (fp == NULL) {
		printf("Couldn't open settings file\n");
		return 0;
	}

	fgets(settings.basename,MAX_FILENAME,fp);
	strip_nl(settings.basename);
	printf("Base Name '%s'\n",settings.basename);

	fgets(settings.runname,MAX_FILENAME,fp);
	strip_nl(settings.runname);
	printf("Run Name '%s'\n",settings.runname);

	fgets(instring,MAX_INPUT,fp);
	if(sscanf(instring,"%d\n",&settings.nruns) != 1) {
		printf("Bad number of files to run\n");
		fclose(fp);
		return 0;
	}

	fgets(instring,MAX_INPUT,fp);
	if(sscanf(instring,"%d",&settings.nfilter) != 1) {
		printf("Bad number of filters \"%s\"\n",instring);
		fclose(fp);
		return 0;
	}

	settings.filters = (LINEAR_FILTER*)malloc(settings.nfilter*sizeof(LINEAR_FILTER));
	for(int i = 0;i<settings.nfilter;i++) {
		fgets(instring,MAX_INPUT,fp);
		strip_nl(instring);
		printf("Loading Filter \"%s\"\n",instring);
		if (!read_settings( instring, &(settings.filters[i]) )) {
			fclose(fp);
			return 0;
		}
	}
	fclose(fp);
	return 1;
}

void clean_datafiles(void) {
	for (int i=0;i<nfiles;i++) {
		remove(datafiles[i]);
	}
}

void add_datafile(char *name, char type, int typesize, int ndim, pos *dims,void *data, int filen) {
	data_id *id;
	id=sdf_create_id(0,name,type,typesize,ndim,dims);
	if (filen < 0) {
		for (int i=0;i<nfiles;i++) {
			sdf_write(datafiles[i],id,data);
		}
	} else {
		sdf_write(datafiles[filen],id,data);
	}
	sdf_free_id(id);
}
