#include "linalg.h"
#include "LinearFilters.h"
#include "EKF.h"
#include "SigmaPoint.h"
#include "Smoother.h"

#include "filtersettings.h"

bool parse_merge1(char *str,SIGMAPOINT_SETTINGS<7+5> *set, float *alpha, unsigned short *nsmooth) {
  char *tstr;
  tstr = str;
  if (!parse_set(tstr,set))
    return false;
  tstr = &(tstr[4]);
  if( sscanf(tstr,"%f %d\n",alpha,nsmooth) != 2){
    return false;
  }
  return true;
}

bool parse_merge2(char *str,SIGMAPOINT_SETTINGS<7+2> *set, float *alpha, unsigned short *nsmooth) {
  char *tstr;
  tstr = str;
  if (!parse_set(tstr,set))
    return false;
  tstr = &(tstr[4]);
  if( sscanf(tstr,"%f %d\n",alpha,nsmooth) != 2){
    return false;
  }
  return true;
}

bool parse_sp(char *str,SIGMAPOINT_SETTINGS<7+5> *set1,SIGMAPOINT_SETTINGS<7+2> *set2, float *alpha1, float *alpha2, unsigned short *nsmooth) {
  char *tstr;
  tstr = str;
  if (!parse_set(tstr,set1))
    return false;
  tstr = &(tstr[4]);
  if (!parse_set(tstr,set2))
    return false;
  tstr = &(tstr[4]);
  if( sscanf(tstr,"%f %f %d\n",alpha1,alpha2,nsmooth) != 3){
    printf("Could not locate the two alphas and smoothing parameter \"%s\"\n",tstr);
    return false;
  }
  return true;
}

bool read_settings(char *str, LINEAR_FILTER *fil) {
  float tempfloat[6];
  if (strncmp(str,"TRUTH_SW",8)==0) {
    fil->filter_type = TRUTH_SW;
    if( sscanf(str,"TRUTH_SW %d\n",&fil->nsmooths) != 1){
      printf("Bad TRUTH_SW \"%s\"\n",str);
      return false;
    }
    return true;
  } else if (strncmp(str,"TRUTH_S_",8)==0) {
    fil->filter_type = TRUTH_S_;
    if( sscanf(str,"TRUTH_S_ %d\n",&fil->nsmooths) != 1){
      printf("Bad TRUTH_S_ \"%s\"\n",str);
      return false;
    }
    return true;
  } else if (strncmp(str,"TRUTH__W",8)==0) {
    fil->filter_type = TRUTH__W;
    if( sscanf(str,"TRUTH__W %d\n",&fil->nsmooths) != 1){
      printf("Bad TRUTH__W \"%s\"\n",str);
      return false;
    }
    return true;
  } else if (strncmp(str,"EKF",3)==0) {
    fil->filter_type = EKF;
    if( sscanf(str,"EKF %d\n",&fil->nsmooths) != 1){
      printf("Bad EKF \"%s\"\n",str);
      return false;
    }
    return true;
  } else if (strncmp(str,"BSEKF",5)==0) {
    fil->filter_type = BSEKF;
    if( sscanf(str,"BSEKF %d\n",&fil->nsmooths) != 1){
      printf("Bad BSEKF \"%s\"\n",str);
      return false;
    }
    return true;
  } else if (strncmp(str,"BLSF1",5)==0) {
    fil->filter_type = BLSF1;
    if( sscanf(str,"BLSF1 %d\n",&fil->nsmooths) != 1){
      printf("Bad BLSF1 \"%s\"\n",str);
      return false;
    }
    return true;
  } else if (strncmp(str,"BLSF2",5)==0) {
    fil->filter_type = BLSF2;
    if( sscanf(str,"BLSF2 %d %f\n",&fil->nsmooths,&tempfloat[0]) != 2){
      printf("Bad BLSF2 \"%s\"\n",str);
      return false;
    }
    fil->alpha[0] = tempfloat[0];
    return true;
  } else if (strncmp(str,"TBLSF2",5)==0) {
    fil->filter_type = TBLSF2;
    if( sscanf(str,"TBLSF2 %d %f\n",&fil->nsmooths,&tempfloat[0]) != 2){
      printf("Bad TBLSF2 \"%s\"\n",str);
      return false;
    }
    fil->alpha[0] = tempfloat[0];
    return true;
  }  else if (strncmp(str,"MERGE1",6)==0) {
    fil->filter_type = MERGE1;
    if ( !parse_merge1(&str[7],&fil->sp_set_prop,&tempfloat[0],&fil->nsmooths) ){
      printf("Bad MERGE1 \"%s\"\n",str);
      return false;
    }
    fil->alpha[0]=tempfloat[0];
    return true;
  } else if (strncmp(str,"MERGE2",6)==0) {
    fil->filter_type = MERGE2;
    if( !parse_merge2(&str[7],&fil->sp_set_meas,&tempfloat[0],&fil->nsmooths)){
      printf("Bad MERGE2 \"%s\"\n",str);
      return false;
    }
    fil->alpha[0] = tempfloat[0];
    return true;
  } else if (strncmp(str,"SIGMA",5)==0) {
    fil->filter_type = SIGMA;
    if( !parse_sp(&str[6],&fil->sp_set_prop,&fil->sp_set_meas,&tempfloat[0],&tempfloat[1],&fil->nsmooths) ){
      printf("Bad SIGMA \"%s\"\n",str);
      return false;
    }
    fil->alpha[0] = tempfloat[0];
    fil->alpha[1] = tempfloat[1];
    return true;
  } else if (strncmp(str,"FP_SIGMA",8)==0) {
    fil->filter_type = FP_SIGMA;
    if( !parse_sp(&str[9],&fil->sp_set_prop,&fil->sp_set_meas,&tempfloat[0],&tempfloat[1],&fil->nsmooths) ){
      printf("Bad FP_SIGMA \"%s\"\n",str);
      return false;
    }
    fil->alpha[0] = tempfloat[0];
    fil->alpha[1] = tempfloat[1];
    return true;
  } else if (strncmp(str,"FP_MERGE1",9)==0) {
    fil->filter_type = FP_MERGE1;
    if ( !parse_merge1(&str[10],&fil->sp_set_prop,&tempfloat[0],&fil->nsmooths) ){
      printf("Bad FP_MERGE1 \"%s\"\n",str);
      return false;
    }
    fil->alpha[0] = tempfloat[0];
    return true;
  } else if (strncmp(str,"FP_MERGE2",9)==0) {
    fil->filter_type = FP_MERGE2;
    NT k,a,b;
    if( !parse_merge2(&str[10],&fil->sp_set_meas,&tempfloat[0],&fil->nsmooths)){
      printf("Bad FP_MERGE2 \"%s\"\n",str);
      return false;
    }
    fil->alpha[0] = tempfloat[0];
    return true;
  } else {
    printf("Bad Filter Type \"%s\"\n",str);
    return false;
  }
}
