#ifndef FILTERSETTINGS_H_
#define FILTERSETTINGS_H_

#include <string.h>
#include <SigmaPoint.h>

enum filter_t {
	TRUTH_SW,		// Use all the the truth to linearize
	TRUTH_S_,		// Use only the truth state
	TRUTH__W,		// Use only the update noise
	EKF,			// Typical Extended Kalman Filter
	BSEKF,			// Backward Smoothing EKF
	BLSF1,			// Batch Least Squares Filter Method 1
	BLSF2,			// Batch Least Squares Filter Method 2
	TBLSF2,			// Truth Based Least Squares Filter Method 2
	MERGE1,			// Merged EKF update Sigma Point measurement
	MERGE2,			// Merged Sigma Point update EKF measurement
	SIGMA,			// 'True' Sigma point
	FP_MERGE1,		// First pass, Merge (EKF update Sigma Point measurement)
	FP_MERGE2,		// First pass, Merge (Sigma update, Merge measurement)
	FP_SIGMA,		// First pass, 'True' Sigma Point
};
struct LINEAR_FILTER {
	filter_t filter_type;			// The type of filter
	char *filename;					// The file name for the results of this filter
	unsigned short nsmooths;		// The number of smoothing passes
	SIGMAPOINT_SETTINGS<7+5> sp_set_prop;
	SIGMAPOINT_SETTINGS<7+2> sp_set_meas;
	float alpha[2];
};

bool read_settings(char*, LINEAR_FILTER*);

template<unsigned short p>
bool parse_set(char *tstr,SIGMAPOINT_SETTINGS<p> *set){
	if (strncmp(tstr,"SIM",3)==0) {
		Generate_Simplex_Set(*set);
  } else if(strncmp(tstr,"O3N",3)==0){
    Generate_Order3_Set(*set);
	} else if(strncmp(tstr,"O5N",3)==0){
		Generate_Order5_Set(*set);
	} else if(strncmp(tstr,"O7N",3)==0){
		Generate_Order7_Set(*set);
  } else if(strncmp(tstr,"O9N",3)==0){
    Generate_Order9_Set(*set);
  } else if(strncmp(tstr,"O5F",3)==0){
    Generate_Order5_Full_Set(*set);
	} else {
		printf("No Sigma Point Set of that type");
		return false;
	}
  return true;
};

#endif
