#ifndef LINEARFILTERS_H_
#define LINEARFILTERS_H_

#include "linalg.h"

template <unsigned short n>
struct GAUSSIAN {
  vec<n>	 mean;	// Mean
  mat<n,n> cov;	// Covariance
  GAUSSIAN() {
  }
  GAUSSIAN(const vec<n> &tm,const mat<n,n> & tc) {
    mean = tm;
    cov = tc;
  }
  GAUSSIAN(const GAUSSIAN<n> &tg) {
    mean = tg.mean;
    cov = tg.cov;
  }
};

template <unsigned short n, unsigned short p, unsigned short t>
struct	MEASUREMENT {
  vec<p>		y;
  mat<p,n>	H;
  mat<p,t>	V;
  bool		havenonlin;
  mat<p,p>	nonlin;
  MEASUREMENT() {
    havenonlin=false;
  }
};

template <unsigned short n, unsigned short p>
struct	ADVANCEMENT {
  vec<n>		x;
  mat<n,n>	 F;
  mat<n,p>	 L;
  bool		havenonlin;
  mat<n,n>	 nonlin;
  ADVANCEMENT() {
    havenonlin = false;
  };
};

template <unsigned short n, unsigned short p>
GAUSSIAN<n> performAdvancement(const GAUSSIAN<n> &curstate, const GAUSSIAN<p> &noise, const ADVANCEMENT<n,p> &adv) {
  GAUSSIAN<n> rval = curstate;
  rval.mean = adv.x + adv.F*rval.mean+adv.L*noise.mean;
  rval.cov = adv.F*rval.cov*trans(adv.F) + adv.L*noise.cov*trans(adv.L);
  if (adv.havenonlin) {
    rval.cov = rval.cov + adv.nonlin;
  }
  return rval;
};

template <unsigned short n, unsigned short p, unsigned short t>
GAUSSIAN<n> performMeasurement(const GAUSSIAN<n> &curstate, const GAUSSIAN<t> &noise, const MEASUREMENT<n,p,t> &meas) {
  GAUSSIAN<n> rval = curstate;
  mat<n,p> K;
  if (meas.havenonlin)
    K = rval.cov * trans(meas.H) * inv( meas.H*rval.cov*trans(meas.H) + meas.V*noise.cov*trans(meas.V) + meas.nonlin);
  else
    K = rval.cov * trans(meas.H) * inv( meas.H*rval.cov*trans(meas.H) + meas.V*noise.cov*trans(meas.V));

  rval.mean = rval.mean - K * (meas.y + meas.H*rval.mean + meas.V*noise.mean);

  mat<n,n> tf = mat<n,n>(1) - K * meas.H;

  if (meas.havenonlin)
    rval.cov = tf*rval.cov*trans(tf) + K * ( meas.V*noise.cov*trans(meas.V) + meas.nonlin)*trans(K);
  else
    rval.cov = tf*rval.cov*trans(tf) + K * ( meas.V*noise.cov*trans(meas.V) )*trans(K);

  return rval;
};

template <unsigned short n,unsigned short p,typename U>
struct ADVPROTO {
  vec<n> (*fun)(mat<n,n>*,mat<n,p>*,const vec<n>&,const vec<p>&,U*);	// The function
  GAUSSIAN<n>	state;
  GAUSSIAN<p> noise;
  U *forcing;
};

template <unsigned short n,unsigned short p, unsigned short t,typename U>
struct MEASPROTO {
  vec<p> (*fun)(mat<p,n>*,mat<p,t>*,const vec<n>&,const vec<t>&,U*);	// The function
  GAUSSIAN<n>	state;
  GAUSSIAN<t> noise;
  U *forcing;
};

#endif /* LINEARFILTERS_H_ */
