#ifndef SMOOTHER_H_
#define SMOOTHER_H_

#include "LinearFilters.h"

template <unsigned short n>
struct DUALSTATE {
	vec<n> state;
	vec<n> dual;
	bool findcov;
	mat<n,n> *cov;
	mat<n,n> *ctf;
};

template <unsigned short n,unsigned short p>
DUALSTATE<n> runBack(const DUALSTATE<n> &currstate,const GAUSSIAN<n> &state,const GAUSSIAN<n> &stateP,const ADVANCEMENT<n,p> &adv,GAUSSIAN<p> *w_var) {
  DUALSTATE<n> prevstate;

  if (w_var != NULL) {
    mat<p,n> tK = w_var->cov * trans(adv.L);
    w_var->mean = w_var->mean + tK * (currstate.dual);
    //It would be good to beable to figure aut the new covariance
  }

  mat<n,n> cov = state.cov + mat<n,n>((NT)1E-16); // The matrix is often very close to singular because of the quaternion
  //mat<n,n> icov = pinv(state.cov);
  mat<n,n> icov = inv(cov);

  mat<n,n> K = stateP.cov*trans(adv.F);

  prevstate.state = stateP.mean + K*currstate.dual;
  //prevstate.dual = pinv(state.cov) * (prevstate.state-state.mean);
  prevstate.dual = icov * (prevstate.state-state.mean);
  //prevstate.dual = lusolve(cov,prevstate.state-state.mean);

  if (currstate.findcov) {
    prevstate.cov = currstate.cov;
    prevstate.ctf = currstate.ctf;
    //cov = stateP.cov + K*((*currstate.ctf)*(*currstate.cov)*trans(*currstate.ctf) - trans(*currstate.ctf) )*trans(K);
    cov = stateP.cov + K*(*currstate.ctf)*(*currstate.cov)*trans(*currstate.ctf)*trans(K) - K*trans(*currstate.ctf)*trans(K);
    cov = (cov + trans(cov))*(0.5);
    *prevstate.cov = cov;
    *prevstate.ctf = icov;
  }
  prevstate.findcov = currstate.findcov;

  return prevstate;
};


template <unsigned short n, unsigned short p, unsigned short t>
vec<t> findN(DUALSTATE<n> &dstate, MEASUREMENT<n,p,t> &meas, GAUSSIAN<t> &n_var) {
	vec<t> v;

	v = -n_var.cov*trans(meas.V)*pinv(meas.V*n_var.cov*trans(meas.V)) *(meas.y + meas.H*dstate.state+meas.V*n_var.mean);

	return v;
};

#endif /* SMOOTHER_H_ */
